# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import os
from os import path
from pathlib import Path

import amazon.ionbenchmark.ion_load_dump as _ion_load_dump

from amazon.ion.simple_types import IonPySymbol


# Global defaults for CLI test specs
_tool_defaults = {
    'iterations': 100,
    'warmups': 0,
    'io_type': 'buffer',
    'command': 'read',
    'api': 'load_dump',
}


class BenchmarkSpec(dict):
    """
    Describes the configuration for a micro benchmark.

    Contains functions for retrieving values that are common to all, and allows dictionary-like access for additional
    parameters.
    
    Benchmark spec files are a stream of one or more Ion structs that describe a benchmark to be run.
    
    Cross-implementation, cross-format fields include `format`, `input_file`, `command`, `api`, `iterations`, `warmups`,
    `name`, and `io_type`.
    Implementation-specific benchmark fields should include a language prefix. For ion-python, the prefix is 'py'.
    Fields that are cross-format but specific to ion-python include `py_c_extension` and `py_gc_disabled`.
    Some data formats have format-specific fields. For example, protobuf has `protobuf_descriptor_file` and 
    `protobuf_type`.
    Finally, there are some fields that are specific to a particular format and implementation, such as 
    `protobuf_py_file` and `protobuf_py_module`.

    The following is a list of all fields supported by ion-python:
     * `name` - a name to use in the test report
     * `format` - the format to use for the benchmark. Can be `ion_text`, `ion_binary`, or another format that is
       supported by an implementation. ion-python supports `json`, `ujson`, `simplejson`, `rapidjson`, `cbor`, `cbor2`,
       `protobuf`, and `self_describing_protobuf`.
       This should really be named something else, like `encoder` since it's a particular implementation for a data
       format that is being tested rather than the data format itself, but it is called `format` so that it is the same
       as the corresponding CLI option for the `read` and `write` commands.
     * `input_file` – the path for a file containing the input data for this benchmark. Can be an absolute path, or the
       path can be relative to the location of the spec file.
     * `command` – one of `read` or `write`
     * `api` – can be `load_dump` or `streaming`
     * `io_type` – can be `buffer` or `file`
     * `iterations` – the total number of iterations the test should take. This does not necessarily correspond to a
       single invocation of the API being tested—rather it is the number of _samples_ that should be measured for the
       benchmark statistics.
     * `warmups` – the number of times the API should be invoked in order to warm up the runtime environment before
       measuring any sample runs.
     * `py_c_extension` – whether the C-extension should be used. (This may be ignored if C extensions are not supported
       by the currently running Python interpreter.) Default: true.
     * `py_gc_disabled` – whether garbage collection should be disabled (paused) for the duration of each benchmark
       sample. Default: false.
     * `protobuf_type` – the name of the protocol buffer type to use from the provided protocol buffer schema
     * `descriptor_file` – A protocol buffer schema in the form of a `FileDescriptorSet` (generated by `protoc`
       using the `--descriptor_set_out` option) that contains the necessary schema for reading/writing the protobuf
       input file.
     * `py_module` – A Python module path pointing to the module, generated by `protoc`, that contains the type
       required to read/write the given input file.
     * `py_file` – A file path pointing to the Python module, generated by `protoc`, that contains the type
       required to read/write the given input file.

    Note—for the `protobuf` format, the `protobuf_type` field and at least one of `descriptor_file`, `py_module`, or
    `py_file` must be provided.
    """
    _data_object = None
    _loader_dumper = None
    _spec_working_directory = None

    def __init__(self, params: dict, user_overrides: dict = None, user_defaults: dict = None, working_directory=None):
        """
        Construct a new BenchmarkSpec, possibly incorporating user supplied defaults or overrides.

        Between the various dicts of parameters, the fields "format", "input_file", "command", "api", "iterations",
        "warmups", and "io_type" must all have a value.

        :param params: Values for this benchmark spec.
        :param user_overrides: Values that override all other values.
        :param user_defaults: Values that override the tool defaults, but not the `params`.
        :param working_directory: reference point to use if `input_file` is a relative path. Defaults to os.getcwd().
        """
        if user_defaults is None:
            user_defaults = {}
        if user_overrides is None:
            user_overrides = {}

        self._spec_working_directory = working_directory or os.getcwd()

        merged = { **_tool_defaults, **user_defaults, **params, **user_overrides }

        # Convert symbols to strings
        for k in merged.keys():
            if isinstance(merged[k], IonPySymbol):
                merged[k] = merged[k].text

        # If not an absolute path, make relative to the working directory.
        input_file = merged['input_file']
        if not path.isabs(input_file):
            input_file = path.join(self._spec_working_directory, input_file)
            merged['input_file'] = input_file

        super().__init__(merged)

        for k in ["format", "input_file", "command", "api", "iterations", "warmups", "io_type"]:
            if self[k] is None:
                raise ValueError(f"Missing required parameter '{k}'")

        if 'name' not in self:
            self['name'] = f'({self.get_format()},{self.derive_operation_name()},{path.basename(self.get_input_file())})'

    def __missing__(self, key):
        # Instead of raising a KeyError like a usual dict, just return None.
        return None

    def get_attribute_as_path(self, key: str):
        """
        Get value from the backing dict, assuming that it is a file path, and appending it to the spec working directory
        if it is a relative path.
        """
        value = self[key]
        if path.isabs(value):
            return value
        else:
            return path.join(self._spec_working_directory, value)

    def get_name(self):
        """
        Get the name of the BenchmarkSpec. If not provided in __init__, one was generated based on the params provided.
        """
        return self["name"]

    def get_format(self):
        return self["format"]

    def get_input_file(self):
        return self["input_file"]

    def get_command(self):
        return self["command"]

    def get_api(self):
        return self["api"]

    def get_io_type(self):
        return self["io_type"]

    def get_iterations(self):
        return self["iterations"]

    def get_warmups(self):
        return self["warmups"]

    def derive_operation_name(self):
        match_arg = [self.get_io_type(), self.get_command(), self.get_api()]
        if match_arg == ['buffer', 'read', 'load_dump']:
            return 'loads'
        elif match_arg == ['buffer', 'write', 'load_dump']:
            return 'dumps'
        elif match_arg == ['file', 'read', 'load_dump']:
            return 'load'
        elif match_arg == ['file', 'write', 'load_dump']:
            return 'dumps'
        else:
            raise NotImplementedError(f"Argument combination not supported: {match_arg}")

    def get_input_file_size(self):
        return Path(self.get_input_file()).stat().st_size

    def get_data_object(self):
        """
        Get the data object to be used for testing. Used for benchmarks that write data.
        """
        if not self._data_object:
            loader = self.get_loader_dumper()
            with open(self.get_input_file(), "rb") as fp:
                self._data_object = loader.load(fp)
        return self._data_object

    def get_loader_dumper(self):
        """
        :return: an object/class/module that has `dump`, `dumps`, `load`, and `loads` for the given test spec.
        """
        if not self._loader_dumper:
            self._loader_dumper = self._get_loader_dumper()
        return self._loader_dumper

    def _get_loader_dumper(self):
        data_format = self.get_format()
        if data_format == 'ion_binary':
            return _ion_load_dump.IonLoadDump(binary=True, c_ext=self['py_c_extension'])
        elif data_format == 'ion_text':
            return _ion_load_dump.IonLoadDump(binary=False, c_ext=self['py_c_extension'])
        elif data_format == 'json':
            import json
            return json
        elif data_format == 'ujson':
            import ujson
            return ujson
        elif data_format == 'simplejson':
            import simplejson
            return simplejson
        elif data_format == 'rapidjson':
            import rapidjson
            return rapidjson
        elif data_format == 'cbor':
            import cbor
            return cbor
        elif data_format == 'cbor2':
            import cbor2
            return cbor2
        elif data_format == 'self_describing_protobuf':
            from self_describing_proto import SelfDescribingProtoSerde
            # TODO: Consider making the cache option configurable from the spec file
            return SelfDescribingProtoSerde(cache_type_info=True)
        elif data_format == 'protobuf':
            import proto
            type_name = self['protobuf_type']
            if not type_name:
                raise ValueError("protobuf format requires the type to be specified")
            if self['py_module']:
                message_type = proto.get_message_type_from_py(type_name, self['py_module'])
            elif self['py_file']:
                message_type = proto.get_message_type_from_py(type_name, "imported_protobuf_module",
                                                              self.get_attribute_as_path('py_file'))
            elif self['descriptor_file']:
                message_type = proto.get_message_type_from_descriptor_set(type_name, self.get_attribute_as_path('descriptor_file'))
            else:
                raise ValueError("format 'protobuf' spec requires py_module, py_file, or descriptor_file")
            return proto.ProtoSerde(message_type)
        else:
            return None



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import gc
import os
import platform
import tempfile
import time
import timeit

from amazon.ionbenchmark.benchmark_spec import BenchmarkSpec
import amazon.ionbenchmark.Format as _format
from amazon.ionbenchmark.sample_dist import SampleDist

_pypy = platform.python_implementation() == 'PyPy'
if not _pypy:
    import tracemalloc


class BenchmarkResult:
    """
    Results generated by the `run_benchmark` function.
    """
    nanos_per_op: SampleDist = None
    ops_per_second: SampleDist = None
    peak_memory_usage = None  # measured in bytes

    def __init__(self, nanos_per_op, ops_per_second, peak_memory_usage):
        self.nanos_per_op = SampleDist(nanos_per_op)
        self.ops_per_second = SampleDist(ops_per_second)
        self.peak_memory_usage = peak_memory_usage


def run_benchmark(benchmark_spec: BenchmarkSpec):
    """
    Run benchmarks for `benchmark_spec`.

    The overall approach of this runner is to time multiple samples where each sample consists of multiple invocations
    of the test function. As a rule of thumb, a sample size of 30 is the minimum needed to have a useful level of
    confidence in the results. The margin of error is (roughly speaking) inversely proportional to the square root of
    the sample size, so adding more samples can increase the confidence, but it will have increasingly diminishing
    improvements. As a rule of thumb, it's never worth having a sample size greater than 1000.

    This approach is sound because of the Central Limit Theorem. For an approachable introduction, see
    https://www.kristakingmath.com/blog/sampling-distribution-of-the-sample-mean.

    The reason for multiple invocations per sample is to prevent very short functions from being dominated by
    differences in memory locations or other small differences from one sample to the next. This runner uses the `Timer`
    utility's `autorange()` function to determine the number of times the function must be invoked for it to run for
    at least 1 second. That number is then used as the number of invocations for _every_ sample in the set.
    """
    test_fun = _create_test_fun(benchmark_spec)

    # memory profiling
    if _pypy:
        peak_memory_usage = None
    else:
        peak_memory_usage = _trace_memory_allocation(test_fun)

    setup = ""
    if benchmark_spec["py_gc_disabled"]:
        setup += "import gc; gc.disable()"
    else:
        setup += "import gc; gc.enable()"

    timer = timeit.Timer(stmt=test_fun, timer=time.perf_counter_ns, setup=setup)

    # warm up
    timer.timeit(benchmark_spec.get_warmups())

    # TODO: Consider making the target batch time or the batch size configurable instead of using this hack.
    if "PYTEST_CURRENT_TEST" in os.environ:
        # make the unit tests run in a reasonable time
        batch_size = 1
    else:
        # range-finding
        # This needs the default timer (measuring in seconds) to work correctly, so it's a different Timer instance.
        (batch_size, _) = timeit.Timer(stmt=test_fun, setup=setup).autorange()
        # Ad hoc testing indicates that samples of 1-2 seconds give tighter results than the default 0.2 seconds, but for
        # very quick testing, this can be annoyingly slow.
        batch_size *= 5  # ~1-2 seconds

    # sample collection (iterations)
    raw_timings = timer.repeat(benchmark_spec.get_iterations(), batch_size)

    # Normalize the samples (i.e. remove the effect of the batch size) before returning the results
    nanos_per_op = [t/batch_size for t in raw_timings]
    ops_per_sec = [1000000000.0 / t for t in nanos_per_op]

    return BenchmarkResult(nanos_per_op, ops_per_sec, peak_memory_usage)


def _create_test_fun(benchmark_spec: BenchmarkSpec):
    """
    Create a benchmark function for the given `benchmark_spec`.
    """
    loader_dumper = benchmark_spec.get_loader_dumper()
    match_arg = [benchmark_spec.get_io_type(), benchmark_spec.get_command(), benchmark_spec.get_api()]

    if match_arg == ['buffer', 'read', 'load_dump']:
        with open(benchmark_spec.get_input_file(), 'rb') as f:
            buffer = f.read()

        def test_fn():
            return loader_dumper.loads(buffer)

    elif match_arg == ['buffer', 'write', 'load_dump']:
        data_obj = benchmark_spec.get_data_object()

        def test_fn():
            return loader_dumper.dumps(data_obj)

    elif match_arg == ['file', 'read', 'load_dump']:
        data_file = benchmark_spec.get_input_file()

        def test_fn():
            with open(data_file, "rb") as f:
                return loader_dumper.load(f)

    elif match_arg == ['file', 'write', 'load_dump']:
        data_obj = benchmark_spec.get_data_object()
        data_format = benchmark_spec.get_format()
        if _format.format_is_binary(data_format) or _format.format_is_ion(data_format):
            def test_fn():
                with tempfile.TemporaryFile(mode="wb") as f:
                    return loader_dumper.dump(data_obj, f)
        else:
            def test_fn():
                with tempfile.TemporaryFile(mode="wt") as f:
                    return loader_dumper.dump(data_obj, f)

    else:
        raise NotImplementedError(f"Argument combination not supported: {match_arg}")

    return test_fn


def _trace_memory_allocation(test_fn, *args, **kwargs):
    """
    Measure the memory allocations in bytes for a single invocation of test_fn
    """
    gc.disable()
    tracemalloc.start()
    test_fn(*args, **kwargs)
    memory_usage_peak = tracemalloc.get_traced_memory()[1]
    tracemalloc.stop()
    gc.enable()
    return memory_usage_peak



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

from functools import partial

from amazon.ion.core import IonEvent, IonEventType, IonType, \
                            ION_STREAM_INCOMPLETE_EVENT, ION_STREAM_END_EVENT, \
                            ION_VERSION_MARKER_EVENT
from amazon.ion.reader import NEXT_EVENT, SKIP_EVENT, read_data_event


e_scalar = partial(IonEvent, IonEventType.SCALAR)
e_null = partial(e_scalar, IonType.NULL)
e_bool = partial(e_scalar, IonType.BOOL)
e_int = partial(e_scalar, IonType.INT)
e_float = partial(e_scalar, IonType.FLOAT)
e_decimal = partial(e_scalar, IonType.DECIMAL)
e_timestamp = partial(e_scalar, IonType.TIMESTAMP)
e_symbol = partial(e_scalar, IonType.SYMBOL)
e_string = partial(e_scalar, IonType.STRING)
e_clob = partial(e_scalar, IonType.CLOB)
e_blob = partial(e_scalar, IonType.BLOB)

e_null_list = partial(e_scalar, IonType.LIST, None)
e_null_sexp = partial(e_scalar, IonType.SEXP, None)
e_null_struct = partial(e_scalar, IonType.STRUCT, None)

e_start = partial(IonEvent, IonEventType.CONTAINER_START)
e_start_list = partial(e_start, IonType.LIST)
e_start_sexp = partial(e_start, IonType.SEXP)
e_start_struct = partial(e_start, IonType.STRUCT)

e_end = partial(IonEvent, IonEventType.CONTAINER_END)
e_end_list = partial(e_end, IonType.LIST)
e_end_sexp = partial(e_end, IonType.SEXP)
e_end_struct = partial(e_end, IonType.STRUCT)

NEXT = NEXT_EVENT
SKIP = SKIP_EVENT
e_read = read_data_event

INC = ION_STREAM_INCOMPLETE_EVENT
END = ION_STREAM_END_EVENT
IVM = ION_VERSION_MARKER_EVENT



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

"""Provides a ``simplejson``-like API for dumping and loading Ion data."""
import io
import warnings
from datetime import datetime
from decimal import Decimal
from io import BytesIO, TextIOBase
from types import GeneratorType

from amazon.ion.reader_text import text_reader
from amazon.ion.writer_text import text_writer
from .core import IonEvent, IonEventType, IonType, ION_STREAM_END_EVENT, Timestamp, ION_VERSION_MARKER_EVENT
from .exceptions import IonException
from .reader import blocking_reader, NEXT_EVENT
from .reader_binary import binary_reader
from .reader_managed import managed_reader
from .simple_types import IonPyList, IonPyDict, IonPyNull, IonPyBool, IonPyInt, IonPyFloat, IonPyDecimal, \
    IonPyTimestamp, IonPyText, IonPyBytes, IonPySymbol, is_null
from .symbols import SymbolToken
from .writer import blocking_writer
from .writer_binary import binary_writer

try:
    import amazon.ion.ionc as ionc
    __IS_C_EXTENSION_SUPPORTED = True
except ModuleNotFoundError:
    __IS_C_EXTENSION_SUPPORTED = False
except ImportError as e:
    __IS_C_EXTENSION_SUPPORTED = False
    warnings.warn(
        f"Failed to load c-extension module: {e.msg} falling back to pure python implementation",
        ImportWarning)

# TODO: when we release a new major version, come up with a better way to encapsulate these two variables.
# __IS_C_EXTENSION_SUPPORTED is a private flag indicating whether the c extension was loaded/is supported.
# c_ext is a user-facing flag to check whether the c extension is available and/or disable the c extension.
# However, if you mutate it, then it can no longer be used to see if the c extension is available.
c_ext = __IS_C_EXTENSION_SUPPORTED

_ION_CONTAINER_END_EVENT = IonEvent(IonEventType.CONTAINER_END)
_IVM = b'\xe0\x01\x00\xea'
_TEXT_TYPES = (TextIOBase, io.StringIO)


def dump_python(obj, fp, imports=None, binary=True, sequence_as_stream=False, skipkeys=False, ensure_ascii=True,
                check_circular=True, allow_nan=True, cls=None, indent=None, separators=None, encoding='utf-8',
                default=None,
                use_decimal=True, namedtuple_as_object=True, tuple_as_array=True, bigint_as_string=False,
                sort_keys=False,
                item_sort_key=None, for_json=None, ignore_nan=False, int_as_string_bitcount=None,
                iterable_as_array=False,
                tuple_as_sexp=False, omit_version_marker=False, **kw):
    """Serialize ``obj`` as an Ion-formatted stream to ``fp`` (a file-like object), using the following conversion
    table::
        +-------------------+-------------------+
        |  Python           |       Ion         |
        |-------------------+-------------------|
        | None              |    null.null      |
        |-------------------+-------------------|
        | IonPyNull(<type>) |    null.<type>    |
        |-------------------+-------------------|
        | True, False,      |                   |
        | IonPyInt(BOOL),   |     bool          |
        | IonPyBool,        |                   |
        |-------------------+-------------------|
        | int (Python 2, 3) |                   |
        | long (Python 2),  |      int          |
        | IonPyInt(INT)     |                   |
        |-------------------+-------------------|
        | float, IonPyFloat |     float         |
        |-------------------+-------------------|
        | Decimal,          |                   |
        | IonPyDecimal      |     decimal       |
        |-------------------+-------------------|
        | datetime,         |                   |
        | Timestamp,        |    timestamp      |
        | IonPyTimestamp    |                   |
        |-------------------+-------------------|
        | SymbolToken,      |                   |
        | IonPySymbol,      |     symbol        |
        | IonPyText(SYMBOL) |                   |
        |-------------------+-------------------|
        | str (Python 3),   |                   |
        | unicode (Python2),|     string        |
        | IonPyText(STRING) |                   |
        |-------------------+-------------------|
        | IonPyBytes(CLOB)  |     clob          |
        |-------------------+-------------------|
        | str (Python 2),   |                   |
        | bytes (Python 3)  |     blob          |
        | IonPyBytes(BLOB)  |                   |
        |-------------------+-------------------|
        | list,             |                   |
        | tuple (when       |                   |
        |  tuple_as_sexp=   |     list          |
        |  False)           |                   |
        | IonPyList(LIST)   |                   |
        |-------------------+-------------------|
        | tuple (when       |                   |
        |  tuple_as_sexp=   |     sexp          |
        |  True),           |                   |
        | IonPyList(SEXP)   |                   |
        |-------------------+-------------------|
        | dict, namedtuple, |                   |
        | IonPyDict         |     struct        |
        +-------------------+-------------------+

    Args:
        obj (Any): A python object to serialize according to the above table. Any Python object which is neither an
            instance of nor inherits from one of the types in the above table will raise TypeError.
        fp (BaseIO): A file-like object.
        imports (Optional[Sequence[SymbolTable]]): A sequence of shared symbol tables to be used by by the writer.
        binary (Optional[True|False]): When True, outputs binary Ion. When false, outputs text Ion.
        sequence_as_stream (Optional[True|False]): When True, if ``obj`` is a sequence, it will be treated as a stream
            of top-level Ion values (i.e. the resulting Ion data will begin with ``obj``'s first element).
            Default: False.
        skipkeys: NOT IMPLEMENTED
        ensure_ascii: NOT IMPLEMENTED
        check_circular: NOT IMPLEMENTED
        allow_nan: NOT IMPLEMENTED
        cls: NOT IMPLEMENTED
        indent (Str): If binary is False and indent is a string, then members of containers will be pretty-printed with
            a newline followed by that string repeated for each level of nesting. None (the default) selects the most
            compact representation without any newlines. Example: to indent with four spaces per level of nesting,
            use ``'    '``.
        separators: NOT IMPLEMENTED
        encoding: NOT IMPLEMENTED
        default: NOT IMPLEMENTED
        use_decimal: NOT IMPLEMENTED
        namedtuple_as_object: NOT IMPLEMENTED
        tuple_as_array: NOT IMPLEMENTED
        bigint_as_string: NOT IMPLEMENTED
        sort_keys: NOT IMPLEMENTED
        item_sort_key: NOT IMPLEMENTED
        for_json: NOT IMPLEMENTED
        ignore_nan: NOT IMPLEMENTED
        int_as_string_bitcount: NOT IMPLEMENTED
        iterable_as_array: NOT IMPLEMENTED
        tuple_as_sexp (Optional[True|False]): When True, all tuple values will be written as Ion s-expressions.
            When False, all tuple values will be written as Ion lists. Default: False.
        omit_version_marker (Optional|True|False): If binary is False and omit_version_marker is True, omits the
            Ion Version Marker ($ion_1_0) from the output.  Default: False.
        **kw: NOT IMPLEMENTED

    """
    raw_writer = binary_writer(imports) if binary else text_writer(indent=indent)
    writer = blocking_writer(raw_writer, fp)
    from_type = _FROM_TYPE_TUPLE_AS_SEXP if tuple_as_sexp else _FROM_TYPE
    if binary or not omit_version_marker:
        writer.send(ION_VERSION_MARKER_EVENT)  # The IVM is emitted automatically in binary; it's optional in text.
    if sequence_as_stream and isinstance(obj, (list, tuple)) or isinstance(obj, GeneratorType):
        # Treat this top-level sequence as a stream; serialize its elements as top-level values, but don't serialize the
        # sequence itself.
        for top_level in obj:
            _dump(top_level, writer, from_type)
    else:
        _dump(obj, writer, from_type)
    writer.send(ION_STREAM_END_EVENT)


_FROM_TYPE = {
    IonPyNull: IonType.NULL,
    type(None): IonType.NULL,
    type(True): IonType.BOOL,
    type(False): IonType.BOOL,
    float: IonType.FLOAT,
    str: IonType.STRING,
    Decimal: IonType.DECIMAL,
    datetime: IonType.TIMESTAMP,
    Timestamp: IonType.TIMESTAMP,
    bytes: IonType.BLOB,
    SymbolToken: IonType.SYMBOL,
    list: IonType.LIST,
    tuple: IonType.LIST,
    dict: IonType.STRUCT,
    int: IonType.INT
}

_FROM_TYPE_TUPLE_AS_SEXP = dict(_FROM_TYPE)
_FROM_TYPE_TUPLE_AS_SEXP.update({
    tuple: IonType.SEXP
})


def _ion_type(obj, from_type):
    types = [type(obj)]
    while types:
        current_type = types.pop()
        if current_type in from_type:
            if current_type is SymbolToken:
                # SymbolToken is a tuple. Since tuple also has a mapping, SymbolToken has to be special-cased
                # to avoid relying on how the dict is ordered.
                return IonType.SYMBOL
            return from_type[current_type]
        types.extend(current_type.__bases__)

    raise TypeError('Unknown scalar type %r' % (type(obj),))


def _dump(obj, writer, from_type, field=None, in_struct=False, depth=0):
    null = is_null(obj)
    try:
        ion_type = obj.ion_type
        ion_nature = True
    except AttributeError:
        ion_type = _ion_type(obj, from_type)
        ion_nature = False
    if ion_type is None:
        raise IonException('Value must have a non-None ion_type: %s, depth: %d, field: %s' % (repr(obj), depth, field))
    if not null and ion_type.is_container:
        if ion_nature:
            event = obj.to_event(IonEventType.CONTAINER_START, field_name=field, in_struct=in_struct, depth=depth)
        else:
            event = IonEvent(IonEventType.CONTAINER_START, ion_type, field_name=field, depth=depth)
        writer.send(event)
        if ion_type is IonType.STRUCT:
            for field, val in iter(obj.items()):
                _dump(val, writer, from_type, field, in_struct=True, depth=depth + 1)
        else:
            for elem in obj:
                _dump(elem, writer, from_type, depth=depth + 1)
        event = _ION_CONTAINER_END_EVENT
    else:
        # obj is a scalar value
        if ion_nature:
            event = obj.to_event(IonEventType.SCALAR, field_name=field, in_struct=in_struct, depth=depth)
        else:
            event = IonEvent(IonEventType.SCALAR, ion_type, obj, field_name=field, depth=depth)
    writer.send(event)


def dumps(obj, imports=None, binary=True, sequence_as_stream=False, skipkeys=False, ensure_ascii=True,
          check_circular=True,
          allow_nan=True, cls=None, indent=None, separators=None, encoding='utf-8', default=None, use_decimal=True,
          namedtuple_as_object=True, tuple_as_array=True, bigint_as_string=False, sort_keys=False, item_sort_key=None,
          for_json=None, ignore_nan=False, int_as_string_bitcount=None, iterable_as_array=False, tuple_as_sexp=False,
          omit_version_marker=False, **kw):
    """Serialize ``obj`` as Python ``string`` or ``bytes`` object, using the conversion table used by ``dump`` (above).
    Common examples are below, please refer to the
    [Ion Cookbook](https://amazon-ion.github.io/ion-docs/guides/cookbook.html) for detailed information.

    Write an Ion string in text format:
        ``simpleion.dumps(ion_object, binary=False)``

    Write an Ion string in binary format:
        ``simpleion.dumps(ion_object, binary=True)``

    Args:
        obj (Any): A python object to serialize according to the above table. Any Python object which is neither an
            instance of nor inherits from one of the types in the above table will raise TypeError.
        imports (Optional[Sequence[SymbolTable]]): A sequence of shared symbol tables to be used by by the writer.
        binary (Optional[True|False]): When True, outputs binary Ion. When false, outputs text Ion.
        sequence_as_stream (Optional[True|False]): When True, if ``obj`` is a sequence, it will be treated as a stream
            of top-level Ion values (i.e. the resulting Ion data will begin with ``obj``'s first element).
            Default: False.
        skipkeys: NOT IMPLEMENTED
        ensure_ascii: NOT IMPLEMENTED
        check_circular: NOT IMPLEMENTED
        allow_nan: NOT IMPLEMENTED
        cls: NOT IMPLEMENTED
        indent (Str): If binary is False and indent is a string, then members of containers will be pretty-printed with
            a newline followed by that string repeated for each level of nesting. None (the default) selects the most
            compact representation without any newlines. Example: to indent with four spaces per level of nesting,
            use ``'    '``.
        separators: NOT IMPLEMENTED
        encoding: NOT IMPLEMENTED
        default: NOT IMPLEMENTED
        use_decimal: NOT IMPLEMENTED
        namedtuple_as_object: NOT IMPLEMENTED
        tuple_as_array: NOT IMPLEMENTED
        bigint_as_string: NOT IMPLEMENTED
        sort_keys: NOT IMPLEMENTED
        item_sort_key: NOT IMPLEMENTED
        for_json: NOT IMPLEMENTED
        ignore_nan: NOT IMPLEMENTED
        int_as_string_bitcount: NOT IMPLEMENTED
        iterable_as_array: NOT IMPLEMENTED
        tuple_as_sexp (Optional[True|False]): When True, all tuple values will be written as Ion s-expressions.
            When False, all tuple values will be written as Ion lists. Default: False.
        omit_version_marker (Optional|True|False): If binary is False and omit_version_marker is True, omits the
            Ion Version Marker ($ion_1_0) from the output.  Default: False.
        **kw: NOT IMPLEMENTED

    Returns:
        Union[str|bytes]: The string or binary representation of the data.  if ``binary=True``, this will be a
            ``bytes`` object, otherwise this will be a ``str`` object (or ``unicode`` in the case of Python 2.x)
    """
    ion_buffer = io.BytesIO()

    dump(obj, ion_buffer, imports=imports, sequence_as_stream=sequence_as_stream, binary=binary, skipkeys=skipkeys,
         ensure_ascii=ensure_ascii, check_circular=check_circular,
         allow_nan=allow_nan, cls=cls, indent=indent, separators=separators, encoding=encoding, default=default,
         use_decimal=use_decimal, namedtuple_as_object=namedtuple_as_object, tuple_as_array=tuple_as_array,
         bigint_as_string=bigint_as_string, sort_keys=sort_keys, item_sort_key=item_sort_key, for_json=for_json,
         ignore_nan=ignore_nan, int_as_string_bitcount=int_as_string_bitcount, iterable_as_array=iterable_as_array,
         tuple_as_sexp=tuple_as_sexp, omit_version_marker=omit_version_marker, **kw)

    ret_val = ion_buffer.getvalue()
    ion_buffer.close()
    if not binary:
        ret_val = ret_val.decode('utf-8')
    return ret_val


def load_python(fp, catalog=None, single_value=True, encoding='utf-8', cls=None, object_hook=None, parse_float=None,
                parse_int=None, parse_constant=None, object_pairs_hook=None, use_decimal=None, parse_eagerly=True,
                **kw):
    """Deserialize ``fp`` (a file-like object), which contains a text or binary Ion stream, to a Python object using the
    following conversion table::
        +-------------------+-------------------+
        |  Ion              |     Python        |
        |-------------------+-------------------|
        | null.<type>       | IonPyNull(<type>) |
        |-------------------+-------------------|
        | bool              |    IonPyBool      |
        |-------------------+-------------------|
        | int               |    IonPyInt       |
        |-------------------+-------------------|
        | float             |    IonPyFloat     |
        |-------------------+-------------------|
        | decimal           |   IonPyDecimal    |
        |-------------------+-------------------|
        | timestamp         |  IonPyTimestamp   |
        |-------------------+-------------------|
        | symbol            |   IonPySymbol     |
        |-------------------+-------------------|
        | string            | IonPyText(STRING) |
        |-------------------+-------------------|
        | clob              |  IonPyBytes(CLOB) |
        |-------------------+-------------------|
        | blob              |  IonPyBytes(BLOB) |
        |-------------------+-------------------|
        | list              |   IonPyList(LIST) |
        |-------------------+-------------------|
        | sexp              |   IonPyList(SEXP) |
        |-------------------+-------------------|
        | struct            |     IonPyDict     |
        +-------------------+-------------------+

    Args:
        fp (BaseIO): A file-like object containing Ion data.
        catalog (Optional[SymbolTableCatalog]): The catalog to use for resolving symbol table imports.
        single_value (Optional[True|False]): When True, the data in ``obj`` is interpreted as a single Ion value, and
            will be returned without an enclosing container. If True and there are multiple top-level values in the Ion
            stream, IonException will be raised. NOTE: this means that when data is dumped using
            ``sequence_as_stream=True``, it must be loaded using ``single_value=False``. Default: True.
        parse_eagerly: (Optional[True|False]) Used in conjunction with ``single_value=False`` to return the result as list
            or an iterator
        encoding: NOT IMPLEMENTED
        cls: NOT IMPLEMENTED
        object_hook: NOT IMPLEMENTED
        parse_float: NOT IMPLEMENTED
        parse_int: NOT IMPLEMENTED
        parse_constant: NOT IMPLEMENTED
        object_pairs_hook: NOT IMPLEMENTED
        use_decimal: NOT IMPLEMENTED
        **kw: NOT IMPLEMENTED

    Returns (Any):
        if single_value is True:
            A Python object representing a single Ion value.
        else:
            A sequence of Python objects representing a stream of Ion values.
    """
    if isinstance(fp, _TEXT_TYPES):
        raw_reader = text_reader(is_unicode=True)
    else:
        pos = fp.tell()
        maybe_ivm = fp.read(4)
        fp.seek(pos)
        if maybe_ivm == _IVM:
            raw_reader = binary_reader()
        else:
            raw_reader = text_reader()
    reader = blocking_reader(managed_reader(raw_reader, catalog), fp)
    if parse_eagerly:
        out = []  # top-level
        _load(out, reader)
        if single_value:
            if len(out) != 1:
                raise IonException('Stream contained %d values; expected a single value.' % (len(out),))
            return out[0]
        return out
    else:
        out = _load_iteratively(reader)
        if single_value:
            result = next(out)
            try:
                next(out)
                raise IonException('Stream contained more than 1 values; expected a single value.')
            except StopIteration:
                return result
        return out


_FROM_ION_TYPE = [
    IonPyNull,
    IonPyBool,
    IonPyInt,
    IonPyFloat,
    IonPyDecimal,
    IonPyTimestamp,
    IonPySymbol,
    IonPyText,
    IonPyBytes,
    IonPyBytes,
    IonPyList,
    IonPyList,
    IonPyDict
]


def _load_iteratively(reader, end_type=IonEventType.STREAM_END):
    event = reader.send(NEXT_EVENT)
    while event.event_type is not end_type:
        ion_type = event.ion_type
        if event.event_type is IonEventType.CONTAINER_START:
            container = _FROM_ION_TYPE[ion_type].from_event(event)
            _load(container, reader, IonEventType.CONTAINER_END, ion_type is IonType.STRUCT)
            yield container
        elif event.event_type is IonEventType.SCALAR:
            if event.value is None or ion_type is IonType.NULL or ion_type.is_container:
                scalar = IonPyNull.from_event(event)
            else:
                scalar = _FROM_ION_TYPE[ion_type].from_event(event)
            yield scalar
        event = reader.send(NEXT_EVENT)


def _load(out, reader, end_type=IonEventType.STREAM_END, in_struct=False):
    def add(obj):
        if in_struct:
            out.add_item(event.field_name.text, obj)
        else:
            out.append(obj)

    event = reader.send(NEXT_EVENT)
    while event.event_type is not end_type:
        ion_type = event.ion_type
        if event.event_type is IonEventType.CONTAINER_START:
            container = _FROM_ION_TYPE[ion_type].from_event(event)
            _load(container, reader, IonEventType.CONTAINER_END, ion_type is IonType.STRUCT)
            add(container)
        elif event.event_type is IonEventType.SCALAR:
            if event.value is None or ion_type is IonType.NULL or event.ion_type.is_container:
                scalar = IonPyNull.from_event(event)
            else:
                scalar = _FROM_ION_TYPE[ion_type].from_event(event)
            add(scalar)
        event = reader.send(NEXT_EVENT)


def loads(ion_str, catalog=None, single_value=True, encoding='utf-8', cls=None, object_hook=None, parse_float=None,
          parse_int=None, parse_constant=None, object_pairs_hook=None, use_decimal=None, parse_eagerly=True,
          text_buffer_size_limit=None, **kw):
    """Deserialize ``ion_str``, which is a string representation of an Ion object, to a Python object using the
    conversion table used by load (above). Common examples are below, please refer to the
    [Ion Cookbook](https://amazon-ion.github.io/ion-docs/guides/cookbook.html) for detailed information.

    Read an Ion string:
        ``simpleion.loads(ion_string)``

    Read an Ion string using an iterator:
        ``
        it = simpleion.loads(ion_string, parse_eagerly=True)
        # iterate through top-level Ion objects
        next(it)
        next(it)
        ``

    Read an Ion string with 50k text_buffer_size_limit:
        ``simpleion.loads(ion_string, text_buffer_size_limit=50000)``

    Args:
        fp (str): A string representation of Ion data.
        catalog (Optional[SymbolTableCatalog]): The catalog to use for resolving symbol table imports.
        single_value (Optional[True|False]): When True, the data in ``ion_str`` is interpreted as a single Ion value,
            and will be returned without an enclosing container. If True and there are multiple top-level values in
            the Ion stream, IonException will be raised. NOTE: this means that when data is dumped using
            ``sequence_as_stream=True``, it must be loaded using ``single_value=False``. Default: True.
        parse_eagerly: (Optional[True|False]) Used in conjunction with ``single_value=False`` to return the result as list
            or an iterator
        text_buffer_size_limit (int): The maximum byte size allowed for text values when the C extension is enabled
            (default: 512 bytes). This option only has an effect when the C extension is enabled (and it is enabled by
            default). When the C extension is disabled, there is no limit on the size of text values.
        encoding: NOT IMPLEMENTED
        cls: NOT IMPLEMENTED
        object_hook: NOT IMPLEMENTED
        parse_float: NOT IMPLEMENTED
        parse_int: NOT IMPLEMENTED
        parse_constant: NOT IMPLEMENTED
        object_pairs_hook: NOT IMPLEMENTED
        use_decimal: NOT IMPLEMENTED
        **kw: NOT IMPLEMENTED

    Returns (Any):
        if single_value is True:
            A Python object representing a single Ion value.
        else:
            A sequence of Python objects representing a stream of Ion values.
    """

    if isinstance(ion_str, bytes):
        ion_buffer = BytesIO(ion_str)
    elif isinstance(ion_str, str):
        ion_buffer = io.StringIO(ion_str)
    else:
        raise TypeError('Unsupported text: %r' % ion_str)

    return load(ion_buffer, catalog=catalog, single_value=single_value, encoding=encoding, cls=cls,
                object_hook=object_hook, parse_float=parse_float, parse_int=parse_int, parse_constant=parse_constant,
                object_pairs_hook=object_pairs_hook, use_decimal=use_decimal, parse_eagerly=parse_eagerly,
                text_buffer_size_limit=text_buffer_size_limit)


def dump_extension(obj, fp, binary=True, sequence_as_stream=False, tuple_as_sexp=False, omit_version_marker=False):
    res = ionc.ionc_write(obj, binary, sequence_as_stream, tuple_as_sexp)

    # TODO support "omit_version_marker" rather than hacking.
    if not binary and not omit_version_marker:
        res = b'$ion_1_0 ' + res
    fp.write(res)


def load_extension(fp, single_value=True, parse_eagerly=True, text_buffer_size_limit=None):
    """
    Args:
        fp (str): A string representation of Ion data.
        single_value (Optional[True|False]): When True, the data in ``ion_str`` is interpreted as a single Ion value,
            and will be returned without an enclosing container. If True and there are multiple top-level values in
            the Ion stream, IonException will be raised. NOTE: this means that when data is dumped using
            ``sequence_as_stream=True``, it must be loaded using ``single_value=False``. Default: True.
        parse_eagerly: (Optional[True|False]) Used in conjunction with ``single_value=False`` to return the result as list
            or an iterator
        text_buffer_size_limit (int): The maximum byte size allowed for text values when the C extension is enabled
            (default: 512 bytes). This option only has an effect when the C extension is enabled (and it is enabled by
            default). When the C extension is disabled, there is no limit on the size of text values.
    """
    iterator = ionc.ionc_read(fp, emit_bare_values=False, text_buffer_size_limit=text_buffer_size_limit)
    if single_value:
        try:
            value = next(iterator)
        except StopIteration:
            return None
        try:
            next(iterator)
            raise IonException('Stream contained more than 1 values; expected a single value.')
        except StopIteration:
            pass
        return value
    if parse_eagerly:
        return list(iterator)
    return iterator


def dump(obj, fp, imports=None, binary=True, sequence_as_stream=False, skipkeys=False, ensure_ascii=True,
         check_circular=True, allow_nan=True, cls=None, indent=None, separators=None, encoding='utf-8', default=None,
         use_decimal=True, namedtuple_as_object=True, tuple_as_array=True, bigint_as_string=False, sort_keys=False,
         item_sort_key=None, for_json=None, ignore_nan=False, int_as_string_bitcount=None, iterable_as_array=False,
         tuple_as_sexp=False, omit_version_marker=False, **kw):
    if c_ext and __IS_C_EXTENSION_SUPPORTED and (imports is None and indent is None):
        return dump_extension(obj, fp, binary=binary, sequence_as_stream=sequence_as_stream,
                              tuple_as_sexp=tuple_as_sexp, omit_version_marker=omit_version_marker)
    else:
        return dump_python(obj, fp, imports=imports, binary=binary, sequence_as_stream=sequence_as_stream,
                           skipkeys=skipkeys, ensure_ascii=ensure_ascii, check_circular=check_circular,
                           allow_nan=allow_nan, cls=cls, indent=indent, separators=separators, encoding=encoding,
                           default=default, use_decimal=use_decimal, namedtuple_as_object=namedtuple_as_object,
                           tuple_as_array=tuple_as_array, bigint_as_string=bigint_as_string, sort_keys=sort_keys,
                           item_sort_key=item_sort_key, for_json=for_json, ignore_nan=ignore_nan,
                           int_as_string_bitcount=int_as_string_bitcount, iterable_as_array=iterable_as_array,
                           tuple_as_sexp=tuple_as_sexp, omit_version_marker=omit_version_marker, **kw)


def load(fp, catalog=None, single_value=True, encoding='utf-8', cls=None, object_hook=None, parse_float=None,
         parse_int=None, parse_constant=None, object_pairs_hook=None, use_decimal=None, parse_eagerly=True,
         text_buffer_size_limit=None, **kw):
    if c_ext and __IS_C_EXTENSION_SUPPORTED and catalog is None:
        return load_extension(fp, parse_eagerly=parse_eagerly, single_value=single_value,
                              text_buffer_size_limit=text_buffer_size_limit)
    else:
        return load_python(fp, catalog=catalog, single_value=single_value, encoding=encoding, cls=cls,
                           object_hook=object_hook, parse_float=parse_float, parse_int=parse_int,
                           parse_constant=parse_constant, object_pairs_hook=object_pairs_hook,
                           use_decimal=use_decimal, parse_eagerly=parse_eagerly, **kw)



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.
from typing import Any, NamedTuple

from pytest import raises

from tests import parametrize, is_exception

import amazon.ion.symbols as symbols

from amazon.ion.exceptions import CannotSubstituteTable

FOO_1_TEXTS = (u'aa', u'bb', u'cc')
FOO_1_TABLE = symbols.shared_symbol_table(
    name=u'foo',
    version=1,
    symbols=FOO_1_TEXTS
)

# Note the one symbol symbol overlap.
FOO_2_TEXTS = (u'cc', u'dd', u'ee')
FOO_2_TABLE = symbols.shared_symbol_table(
    name=u'foo',
    version=2,
    imports=[FOO_1_TABLE],
    symbols=FOO_2_TEXTS
)

# Note the gap in version.
FOO_4_TEXTS = (u'ff', u'gg', u'hh')
FOO_4_TABLE = symbols.shared_symbol_table(
    name=u'foo',
    version=4,
    imports=[FOO_1_TABLE, FOO_2_TABLE],
    symbols=FOO_4_TEXTS
)

REGISTER_TABLES = (
    FOO_1_TABLE, FOO_2_TABLE, FOO_4_TABLE
)


class _P(NamedTuple):
    desc: str
    name: str
    version: int
    max_id: int
    expected: Any

    def __str__(self):
        return '{p.desc} - {p.name}, {p.version}, {p.max_id}'.format(p=self)


@parametrize(
    _P(
        desc='EXACT MATCH',
        name=u'foo',
        version=2,
        max_id=FOO_2_TABLE.max_id,
        expected=FOO_2_TABLE,
    ),
    _P(
        desc='NO MATCH',
        name=u'bar',
        version=1,
        max_id=10,
        expected=symbols.placeholder_symbol_table(name=u'bar', version=1, max_id=10),
    ),
    _P(
        desc='NAME MATCH (HIGHER VERSION)',
        name=u'foo',
        version=5,
        max_id=12,
        expected=symbols.substitute_symbol_table(FOO_4_TABLE, version=5, max_id=12),
    ),
    _P(
        desc='NAME MATCH (LOWER VERSION)',
        name=u'foo',
        version=3,
        max_id=7,
        expected=symbols.substitute_symbol_table(FOO_4_TABLE, version=3, max_id=7),
    ),
    _P(
        # TODO Determine if this is the correct behavior.
        desc='NAME/VERSION MATCH (HIGHER MAX_ID)',
        name=u'foo',
        version=2,
        max_id=15,
        expected=symbols.substitute_symbol_table(FOO_2_TABLE, version=2, max_id=15),
    ),
    _P(
        desc='NAME/VERSION MATCH (LOWER MAX_ID)',
        name=u'foo',
        version=2,
        max_id=4,
        expected=symbols.substitute_symbol_table(FOO_2_TABLE, version=2, max_id=4),
    ),
    _P(
        desc='EXACT MATCH, NO MAX ID',
        name=u'foo',
        version=2,
        max_id=None,
        expected=FOO_2_TABLE,
    ),
    _P(
        desc='NO MATCH, NO MAX ID',
        name=u'bar',
        version=1,
        max_id=None,
        expected=CannotSubstituteTable,
    ),
    _P(
        desc='NAME MATCH, NO MAX ID',
        name=u'foo',
        version=3,
        max_id=None,
        expected=CannotSubstituteTable,
    ),
)
def test_catalog(p):
    catalog = symbols.SymbolTableCatalog()
    for table in REGISTER_TABLES:
        catalog.register(table)

    if is_exception(p.expected):
        with raises(p.expected):
            catalog.resolve(p.name, p.version, p.max_id)
    else:
        resolved = catalog.resolve(p.name, p.version, p.max_id)
        assert p.expected == resolved


class _P(NamedTuple):
    desc: str
    table: symbols.SymbolTable
    expected: Any = ValueError

    def __str__(self):
        return self.desc


@parametrize(
    _P(
        desc='SYSTEM',
        table=symbols.SYSTEM_SYMBOL_TABLE
    ),
    _P(
        desc='LOCAL',
        table=symbols.local_symbol_table()),
    _P(
        desc='PLACEHOLDER',
        table=symbols.placeholder_symbol_table(name=u'placeholder', version=1, max_id=10),
    ),
    _P(
        desc='SUBSTITUTE',
        table=symbols.substitute_symbol_table(FOO_4_TABLE, version=10, max_id=200),
    )
)
def test_catalog_bad_register(p):
    catalog = symbols.SymbolTableCatalog()
    with raises(p.expected):
        catalog.register(p.table)



import shutil
from enum import Enum
import amazon.ion.simpleion as simpleion
import os


def _file_is_ion_binary(file):
    return os.path.splitext(file)[1] == '.10n'


def _file_is_ion_text(file):
    return os.path.splitext(file)[1] == '.ion'


def format_is_ion(format_option):
    return (format_option == Format.ION_BINARY.value) or (format_option == Format.ION_TEXT.value)


def format_is_json(format_option):
    return (format_option == Format.JSON.value) or (format_option == Format.SIMPLEJSON.value) \
        or (format_option == Format.UJSON.value) or (format_option == Format.RAPIDJSON.value)


def format_is_cbor(format_option):
    return (format_option == Format.CBOR.value) or (format_option == Format.CBOR2.value)


def format_is_binary(format_option):
    return format_is_cbor(format_option) or (format_option == Format.ION_BINARY.value) \
           or (format_option == Format.PROTOBUF.value) or (format_option == Format.SD_PROTOBUF.value)


def rewrite_file_to_format(file, format_option):
    temp_file_name_base = 'temp_' + os.path.splitext(os.path.basename(file))[0]
    if format_option == Format.ION_BINARY.value:
        temp_file_name_suffix = '.10n'
    elif format_option == Format.ION_TEXT.value:
        temp_file_name_suffix = '.ion'
    else:
        temp_file_name_suffix = ''
    temp_file_name = temp_file_name_base + temp_file_name_suffix
    # Check the file path
    if os.path.exists(temp_file_name):
        os.remove(temp_file_name)

    if format_is_ion(format_option):
        # Write data if a conversion is required
        if (format_option == Format.ION_BINARY.value and _file_is_ion_text(file)) \
                or (format_option == Format.ION_TEXT.value and _file_is_ion_binary(file)):
            # Load data
            with open(file, 'br') as fp:
                obj = simpleion.load(fp, single_value=False)
            with open(temp_file_name, 'bw') as fp:
                if format_option == Format.ION_BINARY.value:
                    simpleion.dump(obj, fp, binary=True)
                else:
                    simpleion.dump(obj, fp, binary=False)
        else:
            shutil.copy(file, temp_file_name)
    else:
        return file

    return temp_file_name


class Format(Enum):
    """Enumeration of the formats."""
    ION_TEXT = 'ion_text'
    ION_BINARY = 'ion_binary'
    JSON = 'json'
    SIMPLEJSON = 'simplejson'
    UJSON = 'ujson'
    RAPIDJSON = 'rapidjson'
    CBOR = 'cbor'
    CBOR2 = 'cbor2'
    PROTOBUF = 'protobuf'
    SD_PROTOBUF = 'self_describing_protobuf'



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

from functools import partial
from io import BytesIO
from itertools import chain
from datetime import datetime
from decimal import Decimal

from amazon.ion.symbols import SymbolToken
from tests import parametrize
from tests.writer_util import assert_writer_events, generate_scalars, generate_containers, \
                              WriterParameter, SIMPLE_SCALARS_MAP_BINARY, ION_ENCODED_INT_ZERO, VARUINT_END_BYTE

from amazon.ion.core import IonEvent, IonType, IonEventType, timestamp, TimestampPrecision
from amazon.ion.writer import blocking_writer
from amazon.ion.writer_binary_raw import _raw_binary_writer, _write_length
from amazon.ion.writer_buffer import BufferTree

_D = Decimal
_DT = datetime

_E = IonEvent
_IT = IonType
_ET = IonEventType
_P = WriterParameter

_P_FAILURES = [
    _P(
        desc="CONTAINER END AT TOP LEVEL",
        events=[
            _E(_ET.CONTAINER_END)
        ],
        expected=TypeError
    ),
    _P(
        desc="CONTAINER START WITH SCALAR VALUE",
        events=[
            _E(_ET.CONTAINER_START, _IT.BOOL, False)
        ],
        expected=TypeError
    ),
    _P(
        desc="STREAM END BELOW TOP LEVEL",
        events=[
            _E(_ET.CONTAINER_START, _IT.LIST),
            _E(_ET.SCALAR, _IT.STRING, u'foo'),
            _E(_ET.STREAM_END)
        ],
        expected=TypeError
    ),
    _P(
        desc="SCALAR WITH CONTAINER VALUE",
        events=[
            _E(_ET.SCALAR, _IT.STRUCT, u'foo')
        ],
        expected=TypeError
    ),
    _P(
        desc="NULL WITH VALUE",
        events=[
            _E(_ET.SCALAR, _IT.NULL, u'foo')
        ],
        expected=TypeError
    ),
    _P(
        desc="INCORRECT TYPE FOR INT",
        events=[
            _E(_ET.SCALAR, _IT.INT, 1.23e4)
        ],
        expected=TypeError
    ),
    _P(
        desc="INCORRECT TYPE FOR FLOAT",
        events=[
            _E(_ET.SCALAR, _IT.FLOAT, _D(42))
        ],
        expected=TypeError
    ),
    _P(
        desc="INCORRECT TYPE FOR DECIMAL",
        events=[
            _E(_ET.SCALAR, _IT.DECIMAL, 1.23e4)
        ],
        expected=TypeError
    ),
    _P(
        desc="INCORRECT TYPE FOR TIMESTAMP",
        events=[
            _E(_ET.SCALAR, _IT.TIMESTAMP, 123456789)
        ],
        expected=TypeError
    ),
    _P(
        desc="INCORRECT TYPE FOR SYMBOL",
        events=[
            _E(_ET.SCALAR, _IT.SYMBOL, u'symbol')
        ],
        expected=TypeError
    ),
    _P(
        desc="INCORRECT TYPE FOR STRING",
        events=[
            _E(_ET.SCALAR, _IT.STRING, 42)
        ],
        expected=TypeError
    ),
    _P(
        desc="IVM EVENT",  # The IVM is handled by the managed binary writer.
        events=[
            _E(_ET.VERSION_MARKER)
        ],
        expected=TypeError
    )
]

_SIMPLE_CONTAINER_MAP = {
    _IT.LIST: (
        (
            (),
            b'\xB0'
        ),
        (
            (_E(_ET.SCALAR, _IT.INT, 0),),
            bytearray([
                0xB0 | 0x01,  # Int value 0 fits in 1 byte.
                ION_ENCODED_INT_ZERO
            ])
        ),
    ),
    _IT.SEXP: (
        (
            (),
            b'\xC0'
        ),
        (
            (_E(_ET.SCALAR, _IT.INT, 0),),
            bytearray([
                0xC0 | 0x01,  # Int value 0 fits in 1 byte.
                ION_ENCODED_INT_ZERO
            ])
        ),
    ),
    _IT.STRUCT: (
        (
            (),
            b'\xD0'
        ),
        (
            (_E(_ET.SCALAR, _IT.INT, 0, field_name=SymbolToken(None, 10)),),
            bytearray([
                0xDE,  # The lower nibble may vary by implementation. It does not indicate actual length unless it's 0.
                VARUINT_END_BYTE | 2,  # Field name 10 and value 0 each fit in 1 byte.
                VARUINT_END_BYTE | 10,
                ION_ENCODED_INT_ZERO
            ])
        ),
    ),
}


_generate_simple_scalars = partial(generate_scalars, SIMPLE_SCALARS_MAP_BINARY, True)
_generate_simple_containers = partial(generate_containers, _SIMPLE_CONTAINER_MAP, True)


def _generate_annotated_values():
    for value_p in chain(_generate_simple_scalars(), _generate_simple_containers()):
        events = (value_p.events[0].derive_annotations(
            [SymbolToken(None, 10), SymbolToken(None, 11)]),) + value_p.events[1:]
        annot_length = 2  # 10 and 11 each fit in one VarUInt byte
        annot_length_length = 1  # 2 fits in one VarUInt byte
        final_expected = ()
        if isinstance(value_p.expected, (list, tuple)):
            expecteds = value_p.expected
        else:
            expecteds = (value_p.expected,)

        for one_expected in expecteds:
            value_length = len(one_expected)
            length_field = annot_length + annot_length_length + value_length
            wrapper = []
            _write_length(wrapper, length_field, 0xE0)
            wrapper.extend([
                VARUINT_END_BYTE | annot_length,
                VARUINT_END_BYTE | 10,
                VARUINT_END_BYTE | 11
            ])

            exp = bytearray(wrapper) + one_expected
            final_expected += (exp, )

        yield _P(
            desc='ANN %s' % value_p.desc,
            events=events + (_E(_ET.STREAM_END),),
            expected=final_expected,
        )


def new_writer():
    out = BytesIO()
    return out, blocking_writer(_raw_binary_writer(BufferTree()), out)


@parametrize(
    *tuple(chain(
        _P_FAILURES,
        _generate_simple_scalars(),
        _generate_simple_containers(),
        _generate_annotated_values()
    ))
)
def test_raw_writer(p):
    assert_writer_events(p, new_writer)



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

import pytest

from amazon.ion.util import record


def test_default_fields():
    class TestRecord(record('a', 'b', ('c', 5))):
        pass

    a = TestRecord(1, 2)
    assert a.a == 1
    assert a.b == 2
    assert a.c == 5


def test_missing_default():
    with pytest.raises(ValueError):
        class TestRecord(record(('a', 1), 'b')):
            pass


def test_bad_parameter():
    with pytest.raises(ValueError):
        class TestRecord(record(True)):
            pass



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.
from collections import deque
from datetime import timedelta
from decimal import Decimal, localcontext
from enum import IntEnum
from functools import partial
from io import BytesIO
from struct import unpack
from typing import NamedTuple, Optional, Sequence, Callable, Tuple, List

from .core import ION_STREAM_INCOMPLETE_EVENT, ION_STREAM_END_EVENT, ION_VERSION_MARKER_EVENT, \
    IonEventType, IonType, IonEvent, IonThunkEvent, \
    TimestampPrecision, Timestamp, OffsetTZInfo
from .exceptions import IonException
from .sliceable_buffer import SliceableBuffer, IncompleteReadError
from .util import coroutine
from .reader import ReadEventType
from .symbols import SYMBOL_ZERO_TOKEN, SymbolToken


class _TypeID(IntEnum):
    """Type IDs in the binary encoding which is distinct from the :class:`IonType` enum."""
    NULL = 0
    BOOL = 1
    POS_INT = 2
    NEG_INT = 3
    FLOAT = 4
    DECIMAL = 5
    TIMESTAMP = 6
    SYMBOL = 7
    STRING = 8
    CLOB = 9
    BLOB = 10
    LIST = 11
    SEXP = 12
    STRUCT = 13
    ANNOTATION = 14


# Mappings from type code to value type.
_TID_VALUE_TYPE_TABLE = (
    IonType.NULL,
    IonType.BOOL,
    IonType.INT,  # Positive integer
    IonType.INT,  # Negative integer
    IonType.FLOAT,
    IonType.DECIMAL,
    IonType.TIMESTAMP,
    IonType.SYMBOL,
    IonType.STRING,
    IonType.CLOB,
    IonType.BLOB,
    IonType.LIST,
    IonType.SEXP,
    IonType.STRUCT,
    None,  # Annotations do not have an Ion type.
)

# Streams are infinite.
_STREAM_REMAINING = Decimal('Inf')

_VAR_INT_VALUE_MASK = 0b01111111
_VAR_INT_VALUE_BITS = 7
_VAR_INT_SIGN_MASK = 0b01000000
_VAR_INT_SIGN_VALUE_MASK = 0b00111111
_VAR_INT_SIGNAL_MASK = 0b10000000

_SIGNED_INT_SIGN_MASK = 0b10000000
_SIGNED_INT_SIGN_VALUE_MASK = 0b01111111

_LENGTH_LN_MAX = 0xD
_LENGTH_FIELD_FOLLOWS = 0xE
_ALL_LENGTH_LNS = tuple(range(0, _LENGTH_FIELD_FOLLOWS + 1))
_NON_ZERO_LENGTH_LNS = tuple(range(1, _LENGTH_FIELD_FOLLOWS + 1))
_ANNOTATION_LENGTH_LNS = tuple(range(3, _LENGTH_FIELD_FOLLOWS + 1))

_IVM_START_OCTET = 0xE0
_IVM_TAIL = b'\x01\x00\xEA'
_IVM_TAIL_LEN = len(_IVM_TAIL)

# Type IDs for value types that are nullable.
_NULLABLE_TIDS = tuple(range(0, 14))
_NULL_LN = 0xF

_STATIC_SCALARS = (
    # Boolean
    (0x10, IonType.BOOL, False),
    (0x11, IonType.BOOL, True),

    # Zero-values
    (0x20, IonType.INT, 0),
    (0x40, IonType.FLOAT, 0.0),
    (0x50, IonType.DECIMAL, Decimal()),
    (0x70, IonType.SYMBOL, SYMBOL_ZERO_TOKEN),

    # Empty string/clob/blob
    (0x80, IonType.STRING, u''),
    (0x90, IonType.CLOB, b''),
    (0xA0, IonType.BLOB, b''),
)

# Mapping of valid LNs and the struct.unpack format strings
_FLOAT_LN_TABLE = {
    0x4: '>f',
    0x8: '>d'
}

_CONTAINER_TIDS = (_TypeID.LIST, _TypeID.SEXP, _TypeID.STRUCT)


def _gen_type_octet(hn, ln):
    """Generates a type octet from a high nibble and low nibble."""
    return (hn << 4) | ln


def _parse_var_int_components(buf, signed):
    """Parses a ``VarInt`` or ``VarUInt`` field from a file-like object."""
    value = 0
    sign = 1
    while True:
        ch = buf.read(1)
        if ch == '':
            raise IonException('Variable integer under-run')
        octet = ord(ch)
        if signed:
            if octet & _VAR_INT_SIGN_MASK:
                sign = -1
            value = octet & _VAR_INT_SIGN_VALUE_MASK
            signed = False
        else:
            value <<= _VAR_INT_VALUE_BITS
            value |= octet & _VAR_INT_VALUE_MASK

        if octet & _VAR_INT_SIGNAL_MASK:
            break
    return sign, value


def _parse_var_int(buf, signed):
    sign, value = _parse_var_int_components(buf, signed)
    return sign * value


def _parse_signed_int_components(buf):
    """Parses the remainder of a file-like object as a signed magnitude value.

    Returns:
        Returns a pair of the sign bit and the unsigned magnitude.
    """
    sign_bit = 0
    value = 0

    first = True
    while True:
        ch = buf.read(1)
        if ch == b'':
            break
        octet = ord(ch)
        if first:
            if octet & _SIGNED_INT_SIGN_MASK:
                sign_bit = 1
            value = octet & _SIGNED_INT_SIGN_VALUE_MASK
            first = False
        else:
            value <<= 8
            value |= octet

    return sign_bit, value


def _parse_decimal(buf):
    """Parses the remainder of a file-like object as a decimal."""
    exponent = _parse_var_int(buf, signed=True)
    sign_bit, coefficient = _parse_signed_int_components(buf)

    if coefficient == 0:
        # Handle the zero cases--especially negative zero
        value = Decimal((sign_bit, (0,), exponent))
    else:
        coefficient *= sign_bit and -1 or 1
        with localcontext() as context:
            # Adjusting precision for taking into account arbitrarily
            # large/small numbers
            context.prec = len(str(coefficient))
            value = Decimal(coefficient).scaleb(exponent)

    return value


def _parse_sid_iter(data):
    """Parses the given :class:`bytes` data as a list of :class:`SymbolToken`"""
    limit = len(data)
    buf = BytesIO(data)
    while buf.tell() < limit:
        sid = _parse_var_int(buf, signed=False)
        yield SymbolToken(None, sid)


class _ParserContext(NamedTuple):
    buffer: SliceableBuffer
    depth: int


def _invalid_handler(type_octet, ctx):
    """Placeholder handler for invalid type codes."""
    raise IonException(f'Invalid type octet: {type_octet}')


def _var_uint_parser(buffer):
    """
    Parse a Var UInt.

    Return (value, byte_ct) where value is the integer value of the VarUInt
    and byte_ct is the length of the VarUInt.
    """
    value = 0
    while True:
        (octet, buffer) = buffer.read_byte()
        value <<= _VAR_INT_VALUE_BITS
        value |= octet & _VAR_INT_VALUE_MASK
        if octet & _VAR_INT_SIGNAL_MASK:
            break
    return value, buffer


def _var_uint_field_handler(handler, context: _ParserContext):
    """
    Parse the VarUInt length of a field, then delegate to the handler in order
    to parse the field.

    Return the parse result from the field handler.
    """
    (buffer, depth) = context
    length, buffer = _var_uint_parser(buffer)
    return handler(length, _ParserContext(buffer, depth))


def _ivm_handler(context: _ParserContext):
    (buffer, depth) = context
    if depth != 0:
        raise IonException("Ion version markers are only valid at the top-level!")

    (ivm_tail, buffer) = buffer.read_slice(_IVM_TAIL_LEN)
    if _IVM_TAIL != ivm_tail:
        raise IonException('Invalid version marker: %r' % ivm_tail)

    return ION_VERSION_MARKER_EVENT, buffer


def _nop_pad_handler(_, length, context: _ParserContext):
    if not length:
        return None, context.buffer

    (skipped, buffer) = context.buffer.skip(length)
    if skipped < length:
        raise IncompleteReadError("Couldn't complete skip!")

    return None, buffer


def _static_scalar_handler(ion_type, value, context: _ParserContext):
    return IonEvent(IonEventType.SCALAR, ion_type, value, depth=context.depth), context.buffer


def _length_scalar_handler(scalar_factory, ion_type, length, context: _ParserContext):
    """Handles scalars, ``scalar_factory`` is a function that returns a value or thunk."""
    (buffer, depth) = context
    if length == 0:
        data = b''
    else:
        (data, buffer) = buffer.read_slice(length)

    scalar = scalar_factory(data)
    if callable(scalar):
        event = IonThunkEvent(IonEventType.SCALAR, ion_type, scalar, depth=context.depth)
    else:
        event = IonEvent(IonEventType.SCALAR, ion_type, scalar, depth=context.depth)

    return event, buffer


def _annotation_handler(_, length, context: _ParserContext):
    (buffer, depth) = context
    init_size = buffer.size
    anno_length, buffer = _var_uint_parser(buffer)

    if anno_length < 1:
        raise IonException('Invalid annotation length subfield; annotation wrapper must have at least one annotation.')

    (anno_bytes, buffer) = buffer.read_slice(anno_length)
    annotations = tuple(_parse_sid_iter(anno_bytes))

    if length - (init_size - buffer.size) < 1:
        raise IonException("Invalid annotation length subfield; annotation wrapper must wrap non-zero length value.")

    event, buffer = _tlv_parser(_ParserContext(buffer, depth))

    # nop padding comes back as none
    if event is None:
        raise IonException("Cannot annotate nop padding!")

    if event.annotations:
        raise IonException("Cannot nest annotations!")

    actual_length = init_size - buffer.size
    if event.event_type is IonEventType.CONTAINER_START:
        actual_length += event.value
    if actual_length != length:
        raise IonException(f"Expected wrapped value to have length of {length} \
                but was {actual_length}")

    return event.derive_annotations(annotations), buffer


def _ordered_struct_start_handler(length, context: _ParserContext):
    if length < 2:
        raise IonException('Ordered structs (type ID 0xD1) must have at least one field name/value pair.')
    return IonEvent(IonEventType.CONTAINER_START, IonType.STRUCT, value=length, depth=context.depth), context.buffer


def _container_start_handler(ion_type, length, context: _ParserContext):
    # todo: consider extension event to smuggle limit out!
    return IonEvent(IonEventType.CONTAINER_START, ion_type, value=length, depth=context.depth), context.buffer


def _ivm_parser(context: _ParserContext):
    """
    Parse and verify an IVM; used only at start of stream.
    """
    (buffer, depth) = context[0:3]
    (type_octet, buffer) = buffer.read_byte()
    if type_octet != _IVM_START_OCTET:
        raise IonException(
            f'Expected binary version marker, got: {bytes(type_octet)}')

    return _ivm_handler(_ParserContext(buffer, depth))


def _tlv_parser(context: _ParserContext):
    """
    Parse any acceptable top-level or sequence value.

    Validation that IVMs are only at the top-level is in the _ivm_handler.
    """
    (buffer, depth) = context
    (tid, buffer) = buffer.read_byte()
    return _HANDLER_DISPATCH_TABLE[tid](_ParserContext(buffer, depth))


def _struct_item_parser(context: _ParserContext):
    """
    Parse the field and value for an item in a struct.
    """
    (buffer, depth) = context
    field_sid, buffer = _var_uint_parser(buffer)
    event, buffer = _tlv_parser(_ParserContext(buffer, depth))

    if not event:
        return event, buffer
    else:
        return event.derive_field_name(SymbolToken(None, field_sid)), buffer


class _ContextFrame(NamedTuple):
    parser: Callable[[_ParserContext], Tuple[IonEvent, SliceableBuffer]]
    ion_type: Optional[IonType]
    depth: int
    limit: int


@coroutine
def stream_handler():
    """
    Handler for an Ion Binary value-stream.
    """
    buffer = SliceableBuffer.empty()
    # top-level context limit is -1 to denote no limit
    context_stack = deque([_ContextFrame(_tlv_parser, None, 0, -1)])
    cursor = 0
    ion_event = None
    skip_or_next = ReadEventType.NEXT
    expect_data = False
    # will get swapped out for tlv or struct parser in main loop
    parser = _ivm_parser
    parent_type = None
    depth = 0
    limit = -1

    # This is the main event loop for the parser coroutine.
    #
    # Each iteration begins by responding with prior ion_event (None initially)
    # and receiving the user's read event.
    #
    # Then there are two main parts:
    # 1/ handles the users' request, checking invariants and extending the
    #    buffer and/or skipping.
    # 2/ parses if possible then mutates state based on the results of parsing.
    #
    # You should think of them as distinct functions, inlined to avoid stack
    # push/pop overhead.
    while True:
        read_event = yield ion_event
        assert read_event is not None

        # part 1: handle user's read event
        if expect_data:
            if read_event.type is not ReadEventType.DATA:
                raise TypeError("Data expected")
            buffer = buffer.extend(read_event.data)
        else:
            if read_event.type is ReadEventType.DATA:
                raise TypeError("Next or Skip expected")
            skip_or_next = read_event.type

        ion_event = None
        if skip_or_next is ReadEventType.SKIP:
            if parent_type is None:
                raise TypeError("Skip is only allowed within an Ion Container (Struct, List, S-expression)")

            to_skip = limit - cursor
            (skipped, buffer) = buffer.skip(to_skip)
            cursor += skipped
            if cursor < limit:
                ion_event = ION_STREAM_INCOMPLETE_EVENT

        # part 2: parse and reset state

        # loop is to consume but suppress nop padding
        while not ion_event:
            # we might be at end of container
            if cursor == limit:
                ion_event = IonEvent(
                    IonEventType.CONTAINER_END,
                    parent_type,
                    depth=depth - 1)
            # parsing is fun, let's do that!
            else:
                try:
                    (ion_event, new_buff) = parser(_ParserContext(buffer, depth))
                    cursor += buffer.size - new_buff.size
                    buffer = new_buff
                    if 0 < limit < cursor:
                        raise IonException("Passed limit of current container!")
                except IncompleteReadError:
                    if depth == 0 and buffer.size == 0:
                        ion_event = ION_STREAM_END_EVENT
                    else:
                        ion_event = ION_STREAM_INCOMPLETE_EVENT

        event_type = ion_event.event_type
        if event_type.is_stream_signal:
            expect_data = True
        else:
            expect_data = False

            if event_type is IonEventType.CONTAINER_START:
                if ion_event.ion_type is IonType.STRUCT:
                    parser = _struct_item_parser
                else:
                    parser = _tlv_parser
                # we're appropriating "value" for the length of the container
                frame = _ContextFrame(parser, ion_event.ion_type, depth + 1, cursor + ion_event.value)
                context_stack.append(frame)
                ion_event = ion_event.derive_value(None)
            elif event_type is IonEventType.CONTAINER_END:
                context_stack.pop()

            (parser, parent_type, depth, limit) = context_stack[-1]


#
# Scalar Factories
#


def _rslice(data, rem, size):
    start = -rem
    end = start + size
    if end >= 0:
        end = None
    return data[slice(start, end)]


def _int_factory(sign, data):
    def parse_int():
        value = 0
        length = len(data)
        while length >= 8:
            segment = _rslice(data, length, 8)
            value <<= 64
            value |= unpack('>Q', segment)[0]
            length -= 8
        if length >= 4:
            segment = _rslice(data, length, 4)
            value <<= 32
            value |= unpack('>I', segment)[0]
            length -= 4
        if length >= 2:
            segment = _rslice(data, length, 2)
            value <<= 16
            value |= unpack('>H', segment)[0]
            length -= 2
        if length == 1:
            value <<= 8
            value |= data[-length]
        return sign * value
    return parse_int


def _float_factory(data):
    fmt = _FLOAT_LN_TABLE.get(len(data))
    if fmt is None:
        raise ValueError('Invalid data length for float: %d' % len(data))

    return lambda: unpack(fmt, data)[0]


def _decimal_factory(data):
    def parse_decimal():
        return _parse_decimal(BytesIO(data))

    return parse_decimal


def _timestamp_factory(data):
    def parse_timestamp():
        end = len(data)
        buf = BytesIO(data)

        precision = TimestampPrecision.YEAR
        off_sign, off_value = _parse_var_int_components(buf, signed=True)
        off_value *= off_sign
        if off_sign == -1 and off_value == 0:
            # -00:00 (unknown UTC offset) is a naive datetime.
            tz = None
        else:
            tz = OffsetTZInfo(timedelta(minutes=off_value))
        year = _parse_var_int(buf, signed=False)

        if buf.tell() == end:
            month = 1
        else:
            month = _parse_var_int(buf, signed=False)
            precision = TimestampPrecision.MONTH

        if buf.tell() == end:
            day = 1
        else:
            day = _parse_var_int(buf, signed=False)
            precision = TimestampPrecision.DAY

        if buf.tell() == end:
            hour = 0
            minute = 0
        else:
            hour = _parse_var_int(buf, signed=False)
            minute = _parse_var_int(buf, signed=False)
            precision = TimestampPrecision.MINUTE

        if buf.tell() == end:
            second = 0
        else:
            second = _parse_var_int(buf, signed=False)
            precision = TimestampPrecision.SECOND

        if buf.tell() == end:
            fraction = None
        else:
            fraction = _parse_decimal(buf)
            fraction_exponent = fraction.as_tuple().exponent
            if fraction == 0 and fraction_exponent > -1:
                # According to the spec, fractions with coefficients of zero and exponents >= zero are ignored.
                fraction = None

        return Timestamp.adjust_from_utc_fields(
            year, month, day,
            hour, minute, second, None,
            tz,
            precision=precision, fractional_precision=None, fractional_seconds=fraction
        )

    return parse_timestamp


def _symbol_factory(data):
    parse_sid = _int_factory(1, data)

    def parse_symbol():
        sid = parse_sid()
        return SymbolToken(None, sid)

    return parse_symbol


def _string_factory(data):
    return lambda: str(data, 'utf-8')


def _lob_factory(data):
    # Lobs are a trivial return of the byte data.
    return data


#
# Binding Functions
#


# Handler table for type octet to handler function, initialized with the
# invalid handler for all octets.
_HANDLER_DISPATCH_TABLE: List[Callable] = [partial(_invalid_handler, i) for i in range(256)]


def _bind_null_handlers():
    for tid in _NULLABLE_TIDS:
        type_octet = _gen_type_octet(tid, _NULL_LN)
        ion_type = _TID_VALUE_TYPE_TABLE[tid]
        _HANDLER_DISPATCH_TABLE[type_octet] = partial(_static_scalar_handler, ion_type, None)


def _bind_static_scalar_handlers():
    for type_octet, ion_type, value in _STATIC_SCALARS:
        _HANDLER_DISPATCH_TABLE[type_octet] = partial(_static_scalar_handler, ion_type, value)


def _bind_length_handlers(tids, user_handler, lns):
    """Binds a set of handlers with the given factory.

    Args:
        tids (Sequence[int]): The Type IDs to bind to.
        user_handler (Callable): A function that takes as its parameters
            :class:`IonType`, ``length``, and the ``ctx`` context
            returning a ParseResult.
        lns (Sequence[int]): The low-nibble lengths to bind to.
    """
    for tid in tids:
        for ln in lns:
            type_octet = _gen_type_octet(tid, ln)
            ion_type = _TID_VALUE_TYPE_TABLE[tid]
            if ln == 1 and ion_type is IonType.STRUCT:
                handler = partial(_var_uint_field_handler, _ordered_struct_start_handler)
            elif ln < _LENGTH_FIELD_FOLLOWS:
                # Directly partially bind length.
                handler = partial(user_handler, ion_type, ln)
            else:
                # Delegate to length field parsing first.
                handler = partial(_var_uint_field_handler, partial(user_handler, ion_type))
            _HANDLER_DISPATCH_TABLE[type_octet] = handler


def _bind_length_scalar_handlers(tids, scalar_factory, lns=_NON_ZERO_LENGTH_LNS):
    """Binds a set of scalar handlers for an inclusive range of low-nibble values.

    Args:
        tids (Sequence[int]): The Type IDs to bind to.
        scalar_factory (Callable): The factory for the scalar parsing function.
            This function can itself return a function representing a thunk to defer the
            scalar parsing or a direct value.
        lns (Sequence[int]): The low-nibble lengths to bind to.
    """
    handler = partial(_length_scalar_handler, scalar_factory)
    return _bind_length_handlers(tids, handler, lns)


# Populate the actual handlers.
_HANDLER_DISPATCH_TABLE[_IVM_START_OCTET] = _ivm_handler
_bind_null_handlers()
_bind_static_scalar_handlers()
_bind_length_scalar_handlers([_TypeID.POS_INT], partial(_int_factory, 1))
_bind_length_scalar_handlers([_TypeID.NEG_INT], partial(_int_factory, -1))
_bind_length_scalar_handlers([_TypeID.FLOAT], _float_factory, lns=_FLOAT_LN_TABLE.keys())
_bind_length_scalar_handlers([_TypeID.DECIMAL], _decimal_factory)
_bind_length_scalar_handlers([_TypeID.TIMESTAMP], _timestamp_factory)
_bind_length_scalar_handlers([_TypeID.STRING], _string_factory)
_bind_length_scalar_handlers([_TypeID.SYMBOL], _symbol_factory)
_bind_length_scalar_handlers([_TypeID.CLOB, _TypeID.BLOB], _lob_factory)
_bind_length_handlers(_CONTAINER_TIDS, _container_start_handler, _ALL_LENGTH_LNS)
_bind_length_handlers([_TypeID.ANNOTATION], _annotation_handler, _ANNOTATION_LENGTH_LNS)
_bind_length_handlers([_TypeID.NULL], _nop_pad_handler, _ALL_LENGTH_LNS)

# Make immutable.
_HANDLER_DISPATCH_TABLE = tuple(_HANDLER_DISPATCH_TABLE)

binary_reader = stream_handler



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

"""General purpose utilities."""

import sys

from collections import namedtuple
from warnings import warn


class _RecordMetaClass(type):
    """Metaclass for defining named-tuple based immutable record types.

        Deprecated: prefer the built-in NamedTuple
    """

    def __new__(cls, name, bases, attrs):
        warn(f'{cls.__name__} is an internal-only class in ion-python; do not use it for any reason. This '
             f'class is deprecated and may be removed without further warning in any future release. Use `NamedTuple` '
             f'instead.',
             DeprecationWarning, stacklevel=2)

        if attrs.get('_record_sentinel') is None:
            field_declarations = []
            has_record_sentinel = False
            for base_class in bases:
                parent_declarations = getattr(base_class, '_record_fields', None)
                if parent_declarations is not None:
                    field_declarations.extend(parent_declarations)
                    has_record_sentinel = True
            if has_record_sentinel:
                # Only mutate the class if we are directly sub-class a record sentinel.
                names = []
                defaults = []

                has_defaults = False
                for field in field_declarations:
                    if isinstance(field, str):
                        if has_defaults:
                            raise ValueError('Non-defaulted record field must have default: %s' % field)
                        names.append(field)
                    elif isinstance(field, tuple) and len(field) == 2:
                        names.append(field[0])
                        defaults.append(field[1])
                        has_defaults = True
                    else:
                        raise ValueError('Unable to bind record field: %s' % (field,))

                # Construct actual base type/defaults.
                base_class = namedtuple(name, names)
                base_class.__new__.__defaults__ = tuple(defaults)
                # Eliminate our placeholder(s) in the hierarchy.
                bases = (base_class,)

        return super(_RecordMetaClass, cls).__new__(cls, name, bases, attrs)


def record(*fields):
    """Constructs a type that can be extended to create immutable, value types.

    Examples:
        A typical declaration looks like::

            class MyRecord(record('a', ('b', 1))):
                pass

        The above would make a sub-class of ``collections.namedtuple`` that was named ``MyRecord`` with
        a constructor that had the ``b`` field set to 1 by default.

    Note:
        This uses meta-class machinery to rewrite the inheritance hierarchy.
        This is done in order to make sure that the underlying ``namedtuple`` instance is
        bound to the right type name and to make sure that the synthetic class that is generated
        to enable this machinery is not enabled for sub-classes of a user's record class.

    Args:
        fields (list[str | (str, any)]): A sequence of str or pairs that
    """

    class RecordType(object, metaclass=_RecordMetaClass):
        _record_sentinel = True
        _record_fields = fields

    return RecordType


def coroutine(func):
    """Wraps a PEP-342 enhanced generator in a way that avoids boilerplate of the "priming" call to ``next``.

    Args:
        func (Callable): The function constructing a generator to decorate.

    Returns:
        Callable: The decorated generator.
    """

    def wrapper(*args, **kwargs):
        gen = func(*args, **kwargs)
        val = next(gen)
        if val != None:
            raise TypeError('Unexpected value from start of coroutine')
        return gen

    wrapper.__name__ = func.__name__
    wrapper.__doc__ = func.__doc__
    return wrapper


_NON_BMP_OFFSET = 0x10000
_UTF_16_MAX_CODE_POINT = 0xFFFF
_HIGH_SURROGATE_START = 0xD800
_HIGH_SURROGATE_END = 0xDBFF
_LOW_SURROGATE_START = 0xDC00
_LOW_SURROGATE_END = 0xDFFF
_SURROGATE_START = _HIGH_SURROGATE_START
_SURROGATE_END = _LOW_SURROGATE_END


def unicode_iter(val):
    """Provides an iterator over the *code points* of the given Unicode sequence.

    Notes:
        Before PEP-393, Python has the potential to support Unicode as UTF-16 or UTF-32.
        This is reified in the property as ``sys.maxunicode``.  As a result, naive iteration
        of Unicode sequences will render non-character code points such as UTF-16 surrogates.

    Args:
        val (unicode): The unicode sequence to iterate over as integer code points in the range
            ``0x0`` to ``0x10FFFF``.
    """
    val_iter = iter(val)
    while True:
        try:
            code_point = next(_next_code_point(val, val_iter, to_int=ord))
        except StopIteration:
            return
        if code_point is None:
            raise ValueError('Unpaired high surrogate at end of Unicode sequence: %r' % val)
        yield code_point


class CodePoint(int):
    """Evaluates as the ordinal of a code point, while also containing the unicode character representation and
    indicating whether the code point was escaped.
    """

    def __init__(self, *args, **kwargs):
        self.char = None
        self.is_escaped = False


def _next_code_point(val, val_iter, yield_char=False, to_int=lambda x: x):
    """Provides the next *code point* in the given Unicode sequence.

    This generator function yields complete character code points, never incomplete surrogates. When a low surrogate is
    found without following a high surrogate, this function raises ``ValueError`` for having encountered an unpaired
    low surrogate. When the provided iterator ends on a high surrogate, this function yields ``None``. This is the
    **only** case in which this function yields ``None``. When this occurs, the user may append additional data to the
    input unicode sequence and resume iterating through another ``next`` on this generator. When this function receives
    ``next`` after yielding ``None``, it *reinitializes the unicode iterator*. This means that this feature can only
    be used for values that contain an ``__iter__`` implementation that remains at the current position in the data
    when called (e.g. :class:`BufferQueue`). At this point, there are only two possible outcomes:
        * If next code point is a valid low surrogate, this function yields the combined code point represented by the
          surrogate pair.
        * Otherwise, this function raises ``ValueError`` for having encountered an unpaired high surrogate.

    Args:
        val (unicode|BufferQueue): A unicode sequence or unicode BufferQueue over which to iterate.
        val_iter (Iterator[unicode|BufferQueue]): The unicode sequence iterator over ``val`` from which to generate the
            next integer code point in the range ``0x0`` to ``0x10FFFF``.
        yield_char (Optional[bool]): If True **and** the character code point resulted from a surrogate pair, this
            function will yield a :class:`CodePoint` representing the character code point and containing the original
            unicode character. This is useful when the original unicode character will be needed again because UCS2
            Python builds will error when trying to convert code points greater than 0xFFFF back into their
            unicode character representations. This avoids requiring the user to mathematically re-derive the
            surrogate pair in order to successfully convert the code point back to a unicode character.
        to_int (Optional[callable]): A function to call on each element of val_iter to convert that element to an int.
    """
    try:
        high = next(val_iter)
    except StopIteration:
        return
    low = None
    code_point = to_int(high)
    if _LOW_SURROGATE_START <= code_point <= _LOW_SURROGATE_END:
        raise ValueError('Unpaired low surrogate in Unicode sequence: %d' % code_point)
    elif _HIGH_SURROGATE_START <= code_point <= _HIGH_SURROGATE_END:
        def combine_surrogates():
            low_surrogate = next(val_iter)
            low_code_point = to_int(low_surrogate)
            if low_code_point < _LOW_SURROGATE_START or low_code_point > _LOW_SURROGATE_END:
                raise ValueError('Unpaired high surrogate: %d' % code_point)
            # Decode the surrogates
            real_code_point = _NON_BMP_OFFSET
            real_code_point += (code_point - _HIGH_SURROGATE_START) << 10
            real_code_point += (low_code_point - _LOW_SURROGATE_START)
            return real_code_point, low_surrogate

        try:
            code_point, low = combine_surrogates()
        except StopIteration:
            yield None
            val_iter = iter(val)  # More data has appeared in val.
            code_point, low = combine_surrogates()
    if yield_char and low is not None:
        out = CodePoint(code_point)
        if isinstance(val, str):
            # Iterating over a text type returns text types.
            out.char = high + low
        else:
            out.char = chr(high) + chr(low)
    else:
        out = code_point
    yield out


if sys.version_info < (2, 7):
    def bit_length(value):
        if value == 0:
            return 0
        return len(bin(abs(value))) - 2


    def total_seconds(td):
        return (td.microseconds + (td.seconds + td.days * 24 * 3600) * 10 ** 6) / 10 ** 6
else:
    def bit_length(value):
        return value.bit_length()


    def total_seconds(td):
        return td.total_seconds()

bit_length.__doc__ = 'Returns the bit length of an integer'
total_seconds.__doc__ = 'Timedelta ``total_seconds`` with backported support in Python 2.6'


class _EnumMetaClass(type):
    """This is a deprecated, internal-only class in ion-python; do NOT use it for any reason.

    Metaclass for simple enumerations.

    Specifically provides the machinery necessary to emulate simplified Python 3.4 enumerations.
    """

    def __init__(cls, name, bases, attrs):
        members = {}
        # Re-bind any non magic-named method with an instance of the enumeration.
        for attr_name, attr_value in iter(attrs.items()):
            if not attr_name.startswith('_') and not callable(attr_value) and not isinstance(attr_value, property):
                if not isinstance(attr_value, int):
                    raise TypeError('Enum value must be an int: %r' % attr_value)
                actual_value = cls(attr_name, attr_value)
                setattr(cls, attr_name, actual_value)
                members[attr_value] = actual_value

        # Store the members reverse index.
        cls._enum_members = members

        type.__init__(cls, name, bases, attrs)

    def __getitem__(cls, name):
        """Looks up an enumeration value field by integer value."""
        return cls._enum_members[name]

    def __iter__(self):
        """Iterates through the values of the enumeration in no specific order."""
        return iter(self._enum_members.values())


class Enum(int, metaclass=_EnumMetaClass):
    """This is a deprecated, internal-only class in ion-python; do NOT use it for any reason

    Simple integer based enumeration type.

    Examples:
        The typical declaration looks like::

            class MyEnum(Enum):
                A = 1
                B = 2
                C = 3

        At this point ``MyEnum.A`` is an instance of ``MyEnum``.

    Note:
        Proper enumerations were added in Python 3.4 (PEP 435), this is a very simplified implementation
        based loosely on that specification.

        In particular, implicit order of the values is not supported.

    Args:
        value (int): the value associated with the enumeration.

    Attributes:
        name (str): The name of the enum.
        value (int): The original value associated with the enum.
    """
    _enum_members = {}

    def __new__(cls, name, value):
        return int.__new__(cls, value)

    def __init__(self, name, value):
        warn(f'{self.__class__.__name__} is an internal-only class in ion-python; do not use it for any reason. This '
             f'class is deprecated and may be removed without further warning in any future release. Use `IntEnum` '
             f'instead.',
             DeprecationWarning, stacklevel=2)
        super().__init__()
        self.name = name
        self.value = value

    def __init_subclass__(cls, **kwargs):
        warn(f'{cls.__name__} is an internal-only class in ion-python; do not use it for any reason. This '
             f'class is deprecated and may be removed without further warning in any future release. Use `IntEnum` '
             f'instead.',
             DeprecationWarning, stacklevel=2)
        super().__init_subclass__(**kwargs)

    def __getnewargs__(self):
        return self.name, self.value

    def __str__(self):
        return '<%s.%s: %s>' % (type(self).__name__, self.name, self.value)

    __repr__ = __str__




# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

"""Provides common functionality for Ion binary and text readers."""
from collections import deque
from enum import IntEnum

from amazon.ion.symbols import SymbolToken
from .core import DataEvent, IonEventType, Transition
from .core import ION_STREAM_END_EVENT
from .util import coroutine


class CodePoint(int):
    """Evaluates as a code point ordinal, while also containing the unicode character representation and
    indicating whether the code point was escaped.
    """
    def __init__(self, *args, **kwargs):
        self.char = None
        self.is_escaped = False


class CodePointArray:
    """A mutable sequence of code points. Used in place of bytearray() for text values."""
    def __init__(self, initial_bytes=None):
        self.__text = u''
        if initial_bytes is not None:
            for b in initial_bytes:
                self.append(b)

    def append(self, value):
        self.__text += chr(value)

    def extend(self, values):
        if isinstance(values, str):
            self.__text += values
        else:
            assert isinstance(values, bytes)
            for b in iter(values):
                self.append(b)

    def as_symbol(self):
        return SymbolToken(self.__text, sid=None, location=None)

    def as_text(self):
        return self.__text

    def __len__(self):
        return len(self.__text)

    def __repr__(self):
        return 'CodePointArray(text=%s)' % (self.__text,)

    __str__ = __repr__

    def insert(self, index, value):
        raise ValueError('Attempted to add code point in middle of sequence.')

    def __setitem__(self, index, value):
        raise ValueError('Attempted to set code point in middle of sequence.')

    def __getitem__(self, index):
        return self.__text[index]

    def __delitem__(self, index):
        raise ValueError('Attempted to delete from code point sequence.')


_EOF = b'\x04'  # End of transmission character.


class BufferQueue(object):
    """A simple circular buffer of buffers."""
    def __init__(self, is_unicode=False):
        self.__segments = deque()
        self.__offset = 0
        self.__size = 0
        self.__data_cls = CodePointArray if is_unicode else bytearray
        if is_unicode:
            self.__chr = chr
            self.__element_type = str
        else:
            self.__chr = lambda x: x
            self.__element_type = bytes
        self.__ord = ord if is_unicode else lambda x: x
        self.position = 0
        self.is_unicode = is_unicode

    @staticmethod
    def is_eof(c):
        return c is _EOF  # Note reference equality, ensuring that the EOF literal is still illegal as part of the data.

    @staticmethod
    def _incompatible_types(element_type, data):
        raise ValueError('Incompatible input data types. Expected %r, got %r.' % (element_type, type(data)))

    def extend(self, data):
        # TODO Determine if there are any other accumulation strategies that make sense.
        # TODO Determine if we should use memoryview to avoid copying.
        if not isinstance(data, self.__element_type):
            BufferQueue._incompatible_types(self.__element_type, data)
        self.__segments.append(data)
        self.__size += len(data)

    def mark_eof(self):
        self.__segments.append(_EOF)
        self.__size += 1

    def read(self, length, skip=False):
        """Consumes the first ``length`` bytes from the accumulator."""
        if length > self.__size:
            raise IndexError(
                'Cannot pop %d bytes, %d bytes in buffer queue' % (length, self.__size))
        self.position += length
        self.__size -= length
        segments = self.__segments
        offset = self.__offset

        data = self.__data_cls()
        while length > 0:
            segment = segments[0]
            segment_off = offset
            segment_len = len(segment)
            segment_rem = segment_len - segment_off
            segment_read_len = min(segment_rem, length)

            if segment_off == 0 and segment_read_len == segment_rem:
                # consume an entire segment
                if skip:
                    segment_slice = self.__element_type()
                else:
                    segment_slice = segment
            else:
                # Consume a part of the segment.
                if skip:
                    segment_slice = self.__element_type()
                else:
                    segment_slice = segment[segment_off:segment_off + segment_read_len]
                offset = 0
            segment_off += segment_read_len
            if segment_off == segment_len:
                segments.popleft()
                self.__offset = 0
            else:
                self.__offset = segment_off

            if length <= segment_rem and len(data) == 0:
                return segment_slice
            data.extend(segment_slice)
            length -= segment_read_len
        if self.is_unicode:
            return data.as_text()
        else:
            return data

    def read_byte(self):
        if self.__size < 1:
            raise IndexError('Buffer queue is empty')
        segments = self.__segments
        segment = segments[0]
        segment_len = len(segment)
        offset = self.__offset
        if BufferQueue.is_eof(segment):
            octet = _EOF
        else:
            octet = self.__ord(segment[offset])
        offset += 1
        if offset == segment_len:
            offset = 0
            segments.popleft()
        self.__offset = offset
        self.__size -= 1
        self.position += 1
        return octet

    def unread(self, c):
        """Unread the given character, byte, or code point.

        If this is a unicode buffer and the input is an int or byte, it will be interpreted as an ordinal representing
        a unicode code point.

        If this is a binary buffer, the input must be a byte or int; a unicode character will raise an error.
        """
        if self.position < 1:
            raise IndexError('Cannot unread an empty buffer queue.')
        if isinstance(c, str):
            if not self.is_unicode:
                BufferQueue._incompatible_types(self.is_unicode, c)
        else:
            c = self.__chr(c)
        num_code_units = self.is_unicode and len(c) or 1
        if self.__offset == 0:
            if num_code_units == 1:
                if self.is_unicode:
                    segment = c
                else:
                    segment = bytes((c,))
            else:
                segment = c
            self.__segments.appendleft(segment)
        else:
            self.__offset -= num_code_units

            def verify(ch, idx):
                existing = self.__segments[0][self.__offset + idx]
                if existing != ch:
                    raise ValueError('Attempted to unread %s when %s was expected.' % (ch, existing))
            if num_code_units == 1:
                verify(c, 0)
            else:
                for i in range(num_code_units):
                    verify(c[i], i)
        self.__size += num_code_units
        self.position -= num_code_units

    def skip(self, length):
        """Removes ``length`` bytes and returns the number length still required to skip"""
        if length >= self.__size:
            skip_amount = self.__size
            rem = length - skip_amount
            self.__segments.clear()
            self.__offset = 0
            self.__size = 0
            self.position += skip_amount
        else:
            rem = 0
            self.read(length, skip=True)
        return rem

    def __iter__(self):
        while self.__size > 0:
            yield self.read_byte()

    def __len__(self):
        return self.__size


class ReadEventType(IntEnum):
    """Events that are pushed into an Ion reader co-routine.

    Attributes:
        DATA: Indicates more data for the reader.  The expected type is :class:`bytes`.
        NEXT: Indicates that the reader should yield the next event.
        SKIP: Indicates that the reader should proceed to the end of the current container.
            This type is not meaningful at the top-level.
    """
    DATA = 0
    NEXT = 1
    SKIP = 2


NEXT_EVENT = DataEvent(ReadEventType.NEXT, None)
SKIP_EVENT = DataEvent(ReadEventType.SKIP, None)


def read_data_event(data):
    """Simple wrapper over the :class:`DataEvent` constructor to wrap a :class:`bytes` like
    with the ``DATA`` :class:`ReadEventType`.

    Args:
        data (bytes|unicode): The data for the event. Bytes are accepted by both binary and text readers, while unicode
            is accepted by text readers with is_unicode=True.
    """
    return DataEvent(ReadEventType.DATA, data)


@coroutine
def reader_trampoline(start, allow_flush=False):
    """Provides the co-routine trampoline for a reader state machine.

    The given co-routine is a state machine that yields :class:`Transition` and takes
    a Transition of :class:`amazon.ion.core.DataEvent` and the co-routine itself.

    A reader must start with a ``ReadEventType.NEXT`` event to prime the parser.  In many cases
    this will lead to an ``IonEventType.INCOMPLETE`` being yielded, but not always
    (consider a reader over an in-memory data structure).

    Notes:
        A reader delimits its incomplete parse points with ``IonEventType.INCOMPLETE``.
        Readers also delimit complete parse points with ``IonEventType.STREAM_END``;
        this is similar to the ``INCOMPLETE`` case except that it denotes that a logical
        termination of data is *allowed*. When these event are received, the only valid
        input event type is a ``ReadEventType.DATA``.

        Generally, ``ReadEventType.NEXT`` is used to get the next parse event, but
        ``ReadEventType.SKIP`` can be used to skip over the current container.

        An internal state machine co-routine can delimit a state change without yielding
        to the caller by yielding ``None`` event, this will cause the trampoline to invoke
        the transition delegate, immediately.
    Args:
        start: The reader co-routine to initially delegate to.
        allow_flush(Optional[bool]): True if this reader supports receiving ``NEXT`` after
            yielding ``INCOMPLETE`` to trigger an attempt to flush pending parse events,
            otherwise False.

    Yields:
        amazon.ion.core.IonEvent: the result of parsing.

        Receives :class:`DataEvent` to parse into :class:`amazon.ion.core.IonEvent`.
    """
    data_event = yield
    if data_event is None or data_event.type is not ReadEventType.NEXT:
        raise TypeError('Reader must be started with NEXT')
    trans = Transition(None, start)
    while True:
        trans = trans.delegate.send(Transition(data_event, trans.delegate))
        data_event = None
        if trans.event is not None:
            # Only yield if there is an event.
            data_event = (yield trans.event)
            if trans.event.event_type.is_stream_signal:
                if data_event.type is not ReadEventType.DATA:
                    if not allow_flush or not (trans.event.event_type is IonEventType.INCOMPLETE and
                                               data_event.type is ReadEventType.NEXT):
                        raise TypeError('Reader expected data: %r' % (data_event,))
            else:
                if data_event.type is ReadEventType.DATA:
                    raise TypeError('Reader did not expect data')
            if data_event.type is ReadEventType.DATA and len(data_event.data) == 0:
                raise ValueError('Empty data not allowed')
            if trans.event.depth == 0 \
                    and trans.event.event_type is not IonEventType.CONTAINER_START \
                    and data_event.type is ReadEventType.SKIP:
                raise TypeError('Cannot skip at the top-level')


_DEFAULT_BUFFER_SIZE = 8196


@coroutine
def blocking_reader(reader, input, buffer_size=_DEFAULT_BUFFER_SIZE):
    """Provides an implementation of using the reader co-routine with a file-like object.

    Args:
        reader(Coroutine): A reader co-routine.
        input(BaseIO): The file-like object to read from.
        buffer_size(Optional[int]): The optional buffer size to use.
    """
    ion_event = None
    while True:
        read_event = (yield ion_event)
        ion_event = reader.send(read_event)
        while ion_event is not None and ion_event.event_type.is_stream_signal:
            data = input.read(buffer_size)
            if len(data) == 0:
                # End of file.
                if ion_event.event_type is IonEventType.INCOMPLETE:
                    ion_event = reader.send(NEXT_EVENT)
                    continue
                else:
                    yield ION_STREAM_END_EVENT
                    return
            ion_event = reader.send(read_data_event(data))



from typing import NamedTuple, Sequence, Callable

from _pytest.python_api import raises

from amazon.ion.sliceable_buffer import IncompleteReadError, SliceableBuffer
from tests import parametrize

"""
This test rips off the pattern from test_reader_buffer and the relevant
(for now) tests.

The SliceableBuffer is not intended as a drop-in replacement for the
BufferQueue so there are just enough differences to warrant copying and 
modifying vs re-using as-is.

The intention/hope is that eventually that buffer will be replaced by this
one and then this will be the only remaining test.
"""


def read(expected):
    def action(buffer):
        read_len = len(expected)
        (actual, buffer) = buffer.read_slice(read_len)
        assert actual == expected

        return -read_len, buffer

    return action


def expect(exception, action):
    def raises_action(buffer):
        with raises(exception):
            action(buffer)
        return 0, buffer

    return raises_action


def extend(data):
    def action(buffer):
        return len(data), buffer.extend(data)

    return action


def read_byte(expected):
    def action(buffer):
        (actual, buffer) = buffer.read_byte()
        assert expected[0] == actual

        return -1, buffer

    return action


def skip(n, expected_rem=0):
    def action(buffer):
        (skipped, buffer) = buffer.skip(n)
        assert expected_rem == n - skipped
        return -skipped, buffer

    return action


class _P(NamedTuple):
    desc: str
    actions: Sequence[Callable]
    is_unicode: bool = False

    def __str__(self):
        return self.desc


@parametrize(
    _P(
        desc='EMPTY READ BYTE',
        actions=[
            expect(IncompleteReadError, read_byte(b'a')),
        ],
    ),
    _P(
        desc='EMPTY READ SLICE',
        actions=[
            expect(IncompleteReadError, read(b'ignored')),
        ],
    ),
    _P(
        desc='INCOMPLETE READ SLICE',
        actions=[
            extend(b'abcd'),
            read_byte(b'a'),
            expect(IncompleteReadError, read(b'bcde')),
            read_byte(b'b')
        ],
    ),
    _P(
        desc='SINGLE FULL',
        actions=[
            extend(b'abcd'),
            read(b'abcd'),
        ],
    ),
    _P(
        desc='SINGLE BYTE FULL',
        actions=[
            extend(b'a'),
            read_byte(b'a'),
        ],
    ),
    _P(
        desc='CHOMP CHOMP DONE',
        actions=[
            extend(b'ab'),
            read_byte(b'a'),
            read_byte(b'b'),
        ],
    ),
    _P(
        desc='SINGLE SKIP FULL',
        actions=[
            extend(b'abcd'),
            skip(4, expected_rem=0),
            skip(4, expected_rem=4),
        ],
    ),
    _P(
        desc='PARTIAL SKIP INCOMPLETE',
        actions=[
            extend(b'ab'),
            extend(b'cd'),
            skip(5, expected_rem=1),
        ],
    ),
    _P(
        desc='SKIP TWO OF THREE',
        actions=[
            extend(b'ab'),
            extend(b'cd'),
            extend(b'ef'),
            skip(4, expected_rem=0),
            read(b'ef')
        ],
    ),
    _P(
        desc='SINGLE PART',
        actions=[
            extend(b'abcdefg'),
            read(b'ab'),
            read_byte(b'c'),
            skip(2),
            read(b'fg'),
        ],
    ),
    _P(
        desc='MULTI PART AND SPAN',
        actions=[
            extend(b'abcd'),
            extend(b'efgh'),
            read(b'ab'),
            read(b'cdef'),

            extend(b'ijkl'),
            read(b'ghij'),

            extend(b'mnop'),
            read(b'klmnop'),

            extend(b'abcd'),
            extend(b'efgh'),
            extend(b'ijkl'),
            extend(b'm'),
            read(b'abcdefghi'),
            read_byte(b'j'),
            read_byte(b'k'),
            read_byte(b'l'),
            read_byte(b'm'),
        ],
    ),
    _P(
        desc='MULTI MIDDLE',
        actions=[
            extend(b'abcdefg'),
            extend(b'hijklmn'),
            read(b'ab'),
            read(b'cdef'),
            read(b'ghij'),
            read(b'klm'),
            read(b'n')
        ],
    ),
    _P(
        desc='MULTI MIDDLE SKIP',
        actions=[
            extend(b'abcdefg'),
            extend(b'hijklmn'),
            read(b'ab'),
            skip(4),
            read(b'ghij'),
            skip(3),
            read(b'n')
        ],
    ),
)
def test_buffer(p):
    buffer = SliceableBuffer.empty()
    expected_len = 0

    for action in p.actions:
        buffer_len = len(buffer)
        assert buffer_len == expected_len

        len_change, buffer = action(buffer)
        expected_len += len_change

    assert len(buffer) == expected_len



from typing import Sequence, NamedTuple

from amazon.ion.core import Multimap

from tests import parametrize


class _P(NamedTuple):
    pairs: Sequence
    expected_all_values: Sequence
    expected_single_value: Sequence
    expected_total_len: int

    def __str__(self):
        return '{name}'.format(name=self.pairs)


ALL_DATA = _P(
    pairs=[('a', 1), ('a', 2), ('a', 3), ('a', [4, 5, 6]), ('b', 0), ('c', {'x': 'z', 'r': 's'})],
    expected_all_values=[('a', [1, 2, 3, [4, 5, 6]]), ('b', [0]), ('c', [{'x': 'z', 'r': 's'}])],
    expected_single_value=[('a', [4, 5, 6]), ('b', 0), ('c', {'x': 'z', 'r': 's'})],
    expected_total_len=6
)


def _create_multimap_with_items(pairs):
    m = Multimap()
    for pair in pairs:
        m.add_item(pair[0], pair[1])
    return m


@parametrize(
    ALL_DATA
)
def test_add_item(p):
    m = _create_multimap_with_items(p.pairs)
    for expected in p.expected_all_values:
        assert list([x for x in m.get_all_values(expected[0])]) == expected[1]
    for expected in p.expected_single_value:
        assert m[expected[0]] == expected[1]
    assert p.expected_total_len == len(m)


@parametrize(
    (ALL_DATA, ["a"], 2),
    (ALL_DATA, ["b"], 5),
    (ALL_DATA, ["c"], 5),
    (ALL_DATA, ["a", "b"], 1),
    (ALL_DATA, ["a", "b", "c"], 0)
)
def test_delete_item(item):
    m = _create_multimap_with_items(item[0].pairs)
    p, items_to_remove, len_after_removal = item
    for to_remove in items_to_remove:
        del m[to_remove]
    assert len(m) == item[2]


@parametrize(
    {},
    {"a": 1},
    {"a": 1, "b": 2, "c": [1, 2, {3: 4}]}
)
def test_constructor(d):
    m = Multimap(d)
    for k, v in iter(d.items()):
        assert m[k] == v
    assert len(m) == len(d)



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

"""The type mappings for the ``simplejson``-like API.

In particular, this module provides the extension to native Python data types with
particulars of the Ion data model.
"""

from decimal import Decimal

# in Python 3.10, abstract collections have moved into their own module
# for compatibility with 3.10+, first try imports from the new location
# if that fails, try from the pre-3.10 location
try:
    from collections.abc import MutableMapping
except:
    from collections import MutableMapping

from collections import OrderedDict
from decimal import Decimal

from amazon.ion.core import IonType, IonEvent, Timestamp, TIMESTAMP_FRACTIONAL_SECONDS_FIELD, TIMESTAMP_PRECISION_FIELD, \
    TimestampPrecision
from amazon.ion.symbols import SymbolToken


class IonPyNull(object):
    __name__ = 'IonPyNull'
    __qualname__ = 'IonPyNull'

    def __init__(self, ion_type=IonType.NULL, value=None, annotations=()):
        self.ion_type = ion_type
        self.ion_annotations = annotations

    def __bool__(self):
        return False

    def __copy__(self):
        args, kwargs = self._to_constructor_args(self)
        value = self.__class__(*args, **kwargs)
        value.ion_type = self.ion_type
        value.ion_annotations = self.ion_annotations
        return value

    @staticmethod
    def _to_constructor_args(value):
        return (None, value,), {}

    @classmethod
    def from_event(cls, ion_event):
        if ion_event.value is not None:
            args, kwargs = cls._to_constructor_args(ion_event.value)
        else:
            args, kwargs = (None, None, ()), {}
        value = cls(*args, **kwargs)
        value.ion_type = ion_event.ion_type
        value.ion_annotations = ion_event.annotations
        return value

    @classmethod
    def from_value(cls, ion_type, value, annotations=()):
        if value is None:
            value = IonPyNull()
        else:
            args, kwargs = cls._to_constructor_args(value)
            value = cls(*args, **kwargs)
        value.ion_type = ion_type
        value.ion_annotations = annotations
        return value

    def to_event(self, event_type, field_name=None, in_struct=False, depth=None):
        value = self
        if isinstance(self, IonPyNull):
            value = None

        if in_struct:
            if not isinstance(field_name, SymbolToken):
                field_name = SymbolToken(field_name, 0 if field_name is None else None)
        else:
            field_name = None

        return IonEvent(event_type, ion_type=self.ion_type, value=value, field_name=field_name,
                        annotations=self.ion_annotations, depth=depth)


class IonPyDecimal(Decimal):
    __name__ = 'IonPyDecimal'
    __qualname__ = 'IonPyDecimal'
    ion_type = IonType.DECIMAL

    def __new__(cls, *args, **kwargs):
        v = super().__new__(cls, *args, **kwargs)
        v.ion_annotations = ()
        return v

    def __copy__(self):
        args, kwargs = self._to_constructor_args(self)
        value = self.__class__(*args, **kwargs)
        value.ion_type = self.ion_type
        value.ion_annotations = self.ion_annotations
        return value

    @staticmethod
    def _to_constructor_args(value):
        return (value,), {}

    @classmethod
    def from_event(cls, ion_event):
        if ion_event.value is not None:
            args, kwargs = cls._to_constructor_args(ion_event.value)
        else:
            args, kwargs = (None, None, ()), {}
        value = cls(*args, **kwargs)
        value.ion_type = ion_event.ion_type
        value.ion_annotations = ion_event.annotations
        return value

    @classmethod
    def from_value(cls, ion_type, value, annotations=()):
        if value is None:
            value = IonPyNull()
        else:
            args, kwargs = cls._to_constructor_args(value)
            value = cls(*args, **kwargs)
        value.ion_type = ion_type
        value.ion_annotations = annotations
        return value

    def to_event(self, event_type, field_name=None, in_struct=False, depth=None):
        value = self
        if isinstance(self, IonPyNull):
            value = None

        if in_struct:
            if not isinstance(field_name, SymbolToken):
                field_name = SymbolToken(field_name, 0 if field_name is None else None)
        else:
            field_name = None

        return IonEvent(event_type, ion_type=self.ion_type, value=value, field_name=field_name,
                        annotations=self.ion_annotations, depth=depth)


class IonPyBytes(bytes):
    __name__ = 'IonPyBytes'
    __qualname__ = 'IonPyBytes'

    def __new__(cls, *args, **kwargs):
        v = super().__new__(cls, *args, **kwargs)
        v.ion_annotations = ()
        return v

    def __copy__(self):
        args, kwargs = self._to_constructor_args(self)
        value = self.__class__(*args, **kwargs)
        value.ion_type = self.ion_type
        value.ion_annotations = self.ion_annotations
        return value

    @staticmethod
    def _to_constructor_args(value):
        return (value, ), {}

    @classmethod
    def from_event(cls, ion_event):
        if ion_event.value is not None:
            args, kwargs = cls._to_constructor_args(ion_event.value)
        else:
            args, kwargs = (None, None, ()), {}
        value = cls(*args, **kwargs)
        value.ion_type = ion_event.ion_type
        value.ion_annotations = ion_event.annotations
        return value

    @classmethod
    def from_value(cls, ion_type, value, annotations=()):
        if value is None:
            value = IonPyNull()
        else:
            args, kwargs = cls._to_constructor_args(value)
            value = cls(*args, **kwargs)
        value.ion_type = ion_type
        value.ion_annotations = annotations
        return value

    def to_event(self, event_type, field_name=None, in_struct=False, depth=None):
        value = self
        if isinstance(self, IonPyNull):
            value = None

        if in_struct:
            if not isinstance(field_name, SymbolToken):
                field_name = SymbolToken(field_name, 0 if field_name is None else None)
        else:
            field_name = None

        return IonEvent(event_type, ion_type=self.ion_type, value=value, field_name=field_name,
                        annotations=self.ion_annotations, depth=depth)


class IonPyInt(int):
    __name__ = 'IonPyInt'
    __qualname__ = 'IonPyInt'
    ion_type = IonType.INT

    def __new__(cls, *args, **kwargs):
        v = super().__new__(cls, *args, **kwargs)
        v.ion_annotations = ()
        return v

    def __copy__(self):
        args, kwargs = self._to_constructor_args(self)
        value = self.__class__(*args, **kwargs)
        value.ion_type = self.ion_type
        value.ion_annotations = self.ion_annotations
        return value

    @staticmethod
    def _to_constructor_args(value):
        return (value, ), {}

    @classmethod
    def from_event(cls, ion_event):
        if ion_event.value is not None:
            args, kwargs = cls._to_constructor_args(ion_event.value)
        else:
            args, kwargs = (None, None, ()), {}
        value = cls(*args, **kwargs)
        value.ion_type = ion_event.ion_type
        value.ion_annotations = ion_event.annotations
        return value

    @classmethod
    def from_value(cls, ion_type, value, annotations=()):
        if value is None:
            value = IonPyNull()
        else:
            args, kwargs = cls._to_constructor_args(value)
            value = cls(*args, **kwargs)
        value.ion_type = ion_type
        value.ion_annotations = annotations
        return value

    def to_event(self, event_type, field_name=None, in_struct=False, depth=None):
        value = self
        if isinstance(self, IonPyNull):
            value = None

        if in_struct:
            if not isinstance(field_name, SymbolToken):
                field_name = SymbolToken(field_name, 0 if field_name is None else None)
        else:
            field_name = None

        return IonEvent(event_type, ion_type=self.ion_type, value=value, field_name=field_name,
                        annotations=self.ion_annotations, depth=depth)


class IonPyBool(int):
    __name__ = 'IonPyBool'
    __qualname__ = 'IonPyBool'
    ion_type = IonType.BOOL

    def __repr__(self):
        return str(bool(self))

    def __new__(cls, *args, **kwargs):
        v = super().__new__(cls, *args, **kwargs)
        v.ion_annotations = ()
        return v

    def __copy__(self):
        args, kwargs = self._to_constructor_args(self)
        value = self.__class__(*args, **kwargs)
        value.ion_type = self.ion_type
        value.ion_annotations = self.ion_annotations
        return value

    @staticmethod
    def _to_constructor_args(value):
        return (value, ), {}

    @classmethod
    def from_event(cls, ion_event):
        if ion_event.value is not None:
            args, kwargs = cls._to_constructor_args(ion_event.value)
        else:
            args, kwargs = (None, None, ()), {}
        value = cls(*args, **kwargs)
        value.ion_type = ion_event.ion_type
        value.ion_annotations = ion_event.annotations
        return value

    @classmethod
    def from_value(cls, ion_type, value, annotations=()):
        if value is None:
            value = IonPyNull()
        else:
            args, kwargs = cls._to_constructor_args(value)
            value = cls(*args, **kwargs)
        value.ion_type = ion_type
        value.ion_annotations = annotations
        return value

    def to_event(self, event_type, field_name=None, in_struct=False, depth=None):
        value = self
        if isinstance(self, IonPyNull):
            value = None

        if in_struct:
            if not isinstance(field_name, SymbolToken):
                field_name = SymbolToken(field_name, 0 if field_name is None else None)
        else:
            field_name = None

        return IonEvent(event_type, ion_type=self.ion_type, value=value, field_name=field_name,
                        annotations=self.ion_annotations, depth=depth)


class IonPyFloat(float):
    __name__ = 'IonPyFloat'
    __qualname__ = 'IonPyFloat'
    ion_type = IonType.FLOAT

    def __new__(cls, *args, **kwargs):
        v = super().__new__(cls, *args, **kwargs)
        v.ion_annotations = ()
        return v

    def __copy__(self):
        args, kwargs = self._to_constructor_args(self)
        value = self.__class__(*args, **kwargs)
        value.ion_type = self.ion_type
        value.ion_annotations = self.ion_annotations
        return value

    @staticmethod
    def _to_constructor_args(value):
        return (value, ), {}

    @classmethod
    def from_event(cls, ion_event):
        if ion_event.value is not None:
            args, kwargs = cls._to_constructor_args(ion_event.value)
        else:
            args, kwargs = (None, None, ()), {}
        value = cls(*args, **kwargs)
        value.ion_type = ion_event.ion_type
        value.ion_annotations = ion_event.annotations
        return value

    @classmethod
    def from_value(cls, ion_type, value, annotations=()):
        if value is None:
            value = IonPyNull()
        else:
            args, kwargs = cls._to_constructor_args(value)
            value = cls(*args, **kwargs)
        value.ion_type = ion_type
        value.ion_annotations = annotations
        return value

    def to_event(self, event_type, field_name=None, in_struct=False, depth=None):
        value = self
        if isinstance(self, IonPyNull):
            value = None

        if in_struct:
            if not isinstance(field_name, SymbolToken):
                field_name = SymbolToken(field_name, 0 if field_name is None else None)
        else:
            field_name = None

        return IonEvent(event_type, ion_type=self.ion_type, value=value, field_name=field_name,
                        annotations=self.ion_annotations, depth=depth)


class IonPyText(str):
    __name__ = 'IonPyText'
    __qualname__ = 'IonPyText'
    ion_type = IonType.STRING

    def __new__(cls, *args, **kwargs):
        v = super().__new__(cls, *args, **kwargs)
        v.ion_annotations = ()
        return v

    def __copy__(self):
        args, kwargs = self._to_constructor_args(self)
        value = self.__class__(*args, **kwargs)
        value.ion_type = self.ion_type
        value.ion_annotations = self.ion_annotations
        return value

    @staticmethod
    def _to_constructor_args(value):
        return (value, ), {}

    @classmethod
    def from_event(cls, ion_event):
        if ion_event.value is not None:
            args, kwargs = cls._to_constructor_args(ion_event.value)
        else:
            args, kwargs = (None, None, ()), {}
        value = cls(*args, **kwargs)
        value.ion_type = ion_event.ion_type
        value.ion_annotations = ion_event.annotations
        return value

    @classmethod
    def from_value(cls, ion_type, value, annotations=()):
        if value is None:
            value = IonPyNull()
        else:
            args, kwargs = cls._to_constructor_args(value)
            value = cls(*args, **kwargs)
        value.ion_type = ion_type
        value.ion_annotations = annotations
        return value

    def to_event(self, event_type, field_name=None, in_struct=False, depth=None):
        value = self
        if isinstance(self, IonPyNull):
            value = None

        if in_struct:
            if not isinstance(field_name, SymbolToken):
                field_name = SymbolToken(field_name, 0 if field_name is None else None)
        else:
            field_name = None

        return IonEvent(event_type, ion_type=self.ion_type, value=value, field_name=field_name,
                        annotations=self.ion_annotations, depth=depth)


class IonPyTimestamp(Timestamp):
    __name__ = 'IonPyTimestamp'
    __qualname__ = 'IonPyTimestamp'
    ion_type = IonType.TIMESTAMP

    def __new__(cls, *args, **kwargs):
        v = super().__new__(cls, *args, **kwargs)
        v.ion_annotations = ()
        return v

    def __copy__(self):
        args, kwargs = self._to_constructor_args(self)
        value = self.__class__(*args, **kwargs)
        value.ion_type = self.ion_type
        value.ion_annotations = self.ion_annotations
        return value

    @staticmethod
    def _to_constructor_args(ts):
        if isinstance(ts, Timestamp):
            args = (ts.year, ts.month, ts.day, ts.hour, ts.minute, ts.second, None, ts.tzinfo)
            fractional_seconds = getattr(ts, TIMESTAMP_FRACTIONAL_SECONDS_FIELD, None)
            precision = getattr(ts, TIMESTAMP_PRECISION_FIELD, TimestampPrecision.SECOND)
            kwargs = {TIMESTAMP_PRECISION_FIELD: precision, TIMESTAMP_FRACTIONAL_SECONDS_FIELD: fractional_seconds}
        else:
            args = (ts.year, ts.month, ts.day, ts.hour, ts.minute, ts.second, ts.microsecond, ts.tzinfo)
            kwargs = {TIMESTAMP_PRECISION_FIELD: TimestampPrecision.SECOND}
        return args, kwargs

    @classmethod
    def from_event(cls, ion_event):
        if ion_event.value is not None:
            args, kwargs = cls._to_constructor_args(ion_event.value)
        else:
            args, kwargs = (None, None, ()), {}
        value = cls(*args, **kwargs)
        value.ion_type = ion_event.ion_type
        value.ion_annotations = ion_event.annotations
        return value

    @classmethod
    def from_value(cls, ion_type, value, annotations=()):
        if value is None:
            value = IonPyNull()
        else:
            args, kwargs = cls._to_constructor_args(value)
            value = cls(*args, **kwargs)
        value.ion_type = ion_type
        value.ion_annotations = annotations
        return value

    def to_event(self, event_type, field_name=None, in_struct=False, depth=None):
        value = self
        if isinstance(self, IonPyNull):
            value = None

        if in_struct:
            if not isinstance(field_name, SymbolToken):
                field_name = SymbolToken(field_name, 0 if field_name is None else None)
        else:
            field_name = None

        return IonEvent(event_type, ion_type=self.ion_type, value=value, field_name=field_name,
                        annotations=self.ion_annotations, depth=depth)


class IonPySymbol(SymbolToken):
    __name__ = 'IonPySymbol'
    __qualname__ = 'IonPySymbol'
    ion_type = IonType.SYMBOL

    # a good signature: IonPySymbol(ion_type, symbol_token, annotation)
    def __new__(cls, *args, **kwargs):
        v = super().__new__(cls, *args, **kwargs)
        v.ion_annotations = ()
        return v

    def __copy__(self):
        args, kwargs = self._to_constructor_args(self)
        value = self.__class__(*args, **kwargs)
        value.ion_type = self.ion_type
        value.ion_annotations = self.ion_annotations
        return value

    @staticmethod
    def _to_constructor_args(st):
        try:
            args = (st.text, st.sid, st.location)
        except AttributeError:
            args = (st, None, None)
        kwargs = {}
        return args, kwargs

    @classmethod
    def from_event(cls, ion_event):
        if ion_event.value is not None:
            args, kwargs = cls._to_constructor_args(ion_event.value)
        else:
            args, kwargs = (None, None, ()), {}
        value = cls(*args, **kwargs)
        value.ion_type = ion_event.ion_type
        value.ion_annotations = ion_event.annotations
        return value

    @classmethod
    def from_value(cls, ion_type, value, annotations=()):
        if value is None:
            value = IonPyNull()
        else:
            args, kwargs = cls._to_constructor_args(value)
            value = cls(*args, **kwargs)
        value.ion_type = ion_type
        value.ion_annotations = annotations
        return value

    def to_event(self, event_type, field_name=None, in_struct=False, depth=None):
        value = self
        if isinstance(self, IonPyNull):
            value = None

        if in_struct:
            if not isinstance(field_name, SymbolToken):
                field_name = SymbolToken(field_name, 0 if field_name is None else None)
        else:
            field_name = None

        return IonEvent(event_type, ion_type=self.ion_type, value=value, field_name=field_name,
                        annotations=self.ion_annotations, depth=depth)


class IonPyList(list):
    __name__ = 'IonPyList'
    __qualname__ = 'IonPyList'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ion_annotations = ()

    def __copy__(self):
        args, kwargs = self._to_constructor_args(self)
        value = self.__class__(*args, **kwargs)
        value.ion_type = self.ion_type
        value.ion_annotations = self.ion_annotations
        return value

    @staticmethod
    def _to_constructor_args(value):
        return (value, ), {}

    @classmethod
    def from_event(cls, ion_event):
        if ion_event.value is not None:
            args, kwargs = cls._to_constructor_args(ion_event.value)
        else:
            args, kwargs = (), {}
        value = cls(*args, **kwargs)
        value.ion_type = ion_event.ion_type
        value.ion_annotations = ion_event.annotations
        return value

    @classmethod
    def from_value(cls, ion_type, value, annotations=()):
        if value is None:
            value = IonPyNull()
        else:
            args, kwargs = cls._to_constructor_args(value)
            value = cls(*args, **kwargs)
        value.ion_type = ion_type
        value.ion_annotations = annotations
        return value

    def to_event(self, event_type, field_name=None, in_struct=False, depth=None):
        value = None

        if in_struct:
            if not isinstance(field_name, SymbolToken):
                field_name = SymbolToken(field_name, 0 if field_name is None else None)
        else:
            field_name = None

        return IonEvent(event_type, ion_type=self.ion_type, value=value, field_name=field_name,
                        annotations=self.ion_annotations, depth=depth)


class IonPyDict(MutableMapping):
    """
    Dictionary that can hold multiple values for the same key

    In order not to break existing customers, getting and inserting elements with ``[]`` keeps the same behaviour
    as the built-in dict. If multiple elements are already mapped to the key, ``[]`  will return
    the newest one.

    To map multiple elements to a key, use the ``add_item`` operation.
    To retrieve all the values map to a key, use ``get_all_values``.
    """
    __name__ = 'IonPyDict'
    __qualname__ = 'IonPyDict'
    ion_type = IonType.STRUCT

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.ion_annotations = ()
        self.__store = OrderedDict()
        if args is not None and len(args) > 0:
            for key, value in iter(args[0].items()):
                if key in self.__store.keys():
                    self.__store[key].append(value)
                else:
                    self.__store[key] = [value]

    def __getitem__(self, key):
        """
        Return the newest value for the given key. To retrieve all the values map to the key, use ``get_all_values``.
        """
        return self.__store[key][len(self.__store[key]) - 1]  # Return only one in order not to break clients

    def __delitem__(self, key):
        """
        Delete all values for the given key.
        """
        del self.__store[key]

    def __setitem__(self, key, value):
        """
        Set the desired value to the given key.
        """
        self.__store[key] = [value]

    def __len__(self):
        return sum([len(values) for values in iter(self.__store.values())])

    def __iter__(self):
        for key in iter(self.__store.keys()):
            yield key

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return '{%s}' % ', '.join(['%r: %r' % (k, v) for k, v in self.items()])

    def add_item(self, key, value):
        """
        Add a value for the given key. This operation appends the value to the end of the value list instead of
        overwriting the existing value.
        """
        if key in self.__store:
            self.__store[key].append(value)
        else:
            self.__store[key] = [value]

    def get_all_values(self, key):
        """
        Retrieve all the values mapped to the given key
        """
        return self.__store[key]

    def iteritems(self):
        """
        Return an iterator over (key, value) tuple pairs.
        """
        for key in self.__store:
            for value in self.__store[key]:
                yield (key, value)

    def items(self):
        """
        Return a list of the IonPyDict's (key, value) tuple pairs.
        """
        output = []
        for k, v in self.iteritems():
            output.append((k, v))
        return output

    def __copy__(self):
        args, kwargs = self._to_constructor_args(self)
        value = self.__class__(*args, **kwargs)
        value.ion_type = self.ion_type
        value.ion_annotations = self.ion_annotations
        return value

    @staticmethod
    def _to_constructor_args(value):
        return (value, ), {}

    @classmethod
    def from_event(cls, ion_event):
        if ion_event.value is not None:
            args, kwargs = cls._to_constructor_args(ion_event.value)
        else:
            args, kwargs = (), {}
        value = cls(*args, **kwargs)
        value.ion_type = ion_event.ion_type
        value.ion_annotations = ion_event.annotations
        return value

    @classmethod
    def from_value(cls, ion_type, value, annotations=()):
        if value is None:
            value = IonPyNull()
        else:
            args, kwargs = cls._to_constructor_args(value)
            value = cls(*args, **kwargs)
        value.ion_type = ion_type
        value.ion_annotations = annotations
        return value

    def to_event(self, event_type, field_name=None, in_struct=False, depth=None):
        value = None

        if in_struct:
            if not isinstance(field_name, SymbolToken):
                field_name = SymbolToken(field_name, 0 if field_name is None else None)
        else:
            field_name = None

        return IonEvent(event_type, ion_type=self.ion_type, value=value, field_name=field_name,
                        annotations=self.ion_annotations, depth=depth)


def is_null(value):
    """A mechanism to determine if a value is ``None`` or an Ion ``null``."""
    return value is None or isinstance(value, IonPyNull)
